"""
Priors
Ignacio Magana, Rachel Gray, Sergio Vallejo-Peña, Antonio Enea Romano 
"""
from __future__ import absolute_import

import numpy as np

from scipy.interpolate import interp1d

from . import custom_math_priors as _cmp

def pH0(H0, prior='log'):
    """
    Returns p(H0)
    The prior probability of H0

    Parameters
    ----------
    H0 : float or array_like
        Hubble constant value(s) in kms-1Mpc-1
    prior : str, optional
        The choice of prior (default='log')
        if 'log' uses uniform in log prior
        if 'uniform' uses uniform prior

    Returns
    -------
    float or array_like
        p(H0)
    """
    if prior == 'uniform':
        return np.ones(len(H0))
    if prior == 'log':
        return 1./H0

class distance_distribution(object):
    def __init__(self, name):
        self.name = name

        if self.name == 'BBH-powerlaw':
            dist = PriorDict(conversion_function=constrain_m1m2)
            dist['luminosity_distance'] = PowerLaw(alpha=2, minimum=1, maximum=15000)

        if self.name == 'BNS':
            dist = PriorDict(conversion_function=constrain_m1m2)
            dist['luminosity_distance'] = PowerLaw(alpha=2, minimum=1, maximum=1000)

        if self.name == 'NSBH':
            dist = PriorDict(conversion_function=constrain_m1m2)
            dist['luminosity_distance'] = PowerLaw(alpha=2, minimum=1, maximum=1000)

        if self.name == 'BBH-constant':
            dist = PriorDict()
            dist['luminosity_distance'] = PowerLaw(alpha=2, minimum=1, maximum=15000)

        self.dist = dist

    def sample(self, N_samples):
        samples = self.dist.sample(N_samples)
        return samples['luminosity_distance']

    def prob(self, samples):
        return self.dist['luminosity_distance'].prob(samples)

class m_priors(object):
    """
    Parent class with common methods for managing the priors on source frame masses.
    The prior is factorized as :math:`p(m_1,m_2) \\propto p(m_1)p(m_2|m_1)`
    """

    def __init__(self):
        pass

    def update_parameters(self,param_dict):
        """
        Update values of mass priors parameters.
        For more details about functions/parameter definitions see 2111.03604
        """
        #Minimum mass of the PL component of the black hole mass distribution
        if 'mminbh' in param_dict:
            self.mminbh = param_dict['mminbh']
        #Maximum mass of the PL component of the black hole mass distribution
        if 'mmaxbh' in param_dict:
            self.mmaxbh = param_dict['mmaxbh']
        #Spectral index for the PL of the primary mass distribution
        if 'alpha' in param_dict:
            self.alpha = param_dict['alpha']
        #Mean of the Gaussian component in the primary mass distribution
        if 'mu_g' in param_dict:
            self.mu_g = param_dict['mu_g']
        #Width of the Gaussian component in the primary mass distribution
        if 'sigma_g' in param_dict:
            self.sigma_g = param_dict['sigma_g']
        #Fraction of the model in the Gaussian component
        if 'lambda_peak' in param_dict:
            self.lambda_peak = param_dict['lambda_peak']
        #PL slope of the primary mass distribution for masses below mbreak
        if 'alpha_1' in param_dict:
            self.alpha_1 = param_dict['alpha_1']
        #PL slope for the primary mass distribution for masses above mbreak
        if 'alpha_2' in param_dict:
            self.alpha_2 = param_dict['alpha_2']
        #The fraction of the way between mminbh and mmaxbh at which the primary mass distribution breaks
        if 'b' in param_dict:
            self.b = param_dict['b']
        #Range of mass tapering on the lower end of the mass distribution
        if 'delta_m' in param_dict:
            self.delta_m = param_dict['delta_m']
        #Spectral index for the PL of the mass ratio distribution
        if 'beta' in param_dict:
            self.beta = param_dict['beta']        
        #Minimum mass of the neutron star distribution
        if 'mminns' in param_dict:
            self.mminns = param_dict['mminns']
        #Maximum mass of the neutron star distribution
        if 'mmaxns' in param_dict:
            self.mmaxns = param_dict['mmaxns']
        #Spectral index for the PL of the neutron star mass distribution
        if 'alphans' in param_dict:
            self.alphans = param_dict['alphans']
        #updates the mass prior objects dictionary and some auxiliary variables of the mass children classes
        self.update_mass_priors()

    def joint_prob(self, ms1, ms2):
        """
        This method returns the joint probability :math:`p(m_1,m_2)`

        Parameters
        ----------
        ms1: np.array(matrix)
            mass one in solar masses
        ms2: dict
            mass two in solar masses
        """

        to_ret = self.mdis['mass_1'].prob(ms1)*self.mdis['mass_2'].conditioned_prob(ms2,self.mmin*np.ones_like(ms1),np.minimum(ms1,self.mmax2))
        #print('ms1',ms1)
        #print('ms2',ms2)
        #print('mmin',self.mmin)
        #print('mmax2',self.mmax2)
        #print('to_ret',to_ret)

        return to_ret
    
    def log_joint_prob(self,ms1, ms2):
        
        to_ret = np.log(self.joint_prob(ms1, ms2))
        to_ret[np.isnan(to_ret)] = -np.inf

        return to_ret

    def sample(self, Nsample):
        """
        *Not used in O4, due to the use of injections instead of Pdet*
        This method samples from the joint probability :math:`p(m_1,m_2)`

        Parameters
        ----------
        Nsample: int
            Number of samples you want
        """

        vals_m1 = np.random.rand(Nsample)
        vals_m2 = np.random.rand(Nsample)

        m1_trials = np.logspace(np.log10(self.mdis['mass_1'].minimum),np.log10(self.mdis['mass_1'].maximum),10000)
        m2_trials = np.logspace(np.log10(self.mdis['mass_2'].minimum),np.log10(self.mdis['mass_2'].maximum),10000)

        cdf_m1_trials = self.mdis['mass_1'].cdf(m1_trials)
        cdf_m2_trials = self.mdis['mass_2'].cdf(m2_trials)

        m1_trials = np.log10(m1_trials)
        m2_trials = np.log10(m2_trials)

        _,indxm1 = np.unique(cdf_m1_trials,return_index=True)
        _,indxm2 = np.unique(cdf_m2_trials,return_index=True)

        interpo_icdf_m1 = interp1d(cdf_m1_trials[indxm1],m1_trials[indxm1],bounds_error=False,fill_value=(m1_trials[0],m1_trials[-1]))
        interpo_icdf_m2 = interp1d(cdf_m2_trials[indxm2],m2_trials[indxm2],bounds_error=False,fill_value=(m2_trials[0],m2_trials[-1]))

        mass_1_samples = 10**interpo_icdf_m1(vals_m1)
        mass_2_samples = 10**interpo_icdf_m2(vals_m2*self.mdis['mass_2'].cdf(mass_1_samples))

        return mass_1_samples, mass_2_samples

class BBH_powerlaw(m_priors):
    """
    Child class for BBH power law distribution.
    
    Parameters
    -------------
    mminbh: Minimum mass of the PL component of the black hole mass distribution
    mmaxbh: Maximum mass of the PL component of the black hole mass distribution
    alpha: Spectral index for the PL of the primary mass distribution
    beta: Spectral index for the PL of the mass ratio distribution

    The default values of the parameters are set to the corresponding median values in the uniform priors reported in 2111.03604

    ************
    NOTE: The spectral indices passed to PowerLaw_math are alpha=-self.alpha, and alpha=self.beta, according to eqs. A8,A10 in 2111.03604
    ************
    
    The method m_priors.update_parameters is used  in the constructor to initialize the objects
    """
    def __init__(self,mminbh=6.0,mmaxbh=125.0,alpha=6.75,beta=4.0):
        super().__init__()

        self.update_parameters(param_dict={'alpha':alpha, 'beta':beta, 'mminbh':mminbh, 'mmaxbh':mmaxbh})               
              
    def update_mass_priors(self):
        ''' 
        This method creates a dictionary of mass distributions objects. 
        It sets the maximum value of the primary mass distribution mmax to mmaxbh, 
        the minimum value of the secondary mass distribution mmin to mminbh, 
        and the maximum value of the secondary mass distribution mmax2 to mmaxbh.
        It's called by update_paratemters everytime the mass priors parameters are changed.
        Every mass priors model has a different implementation because the distributions are different,
        and mmax, mmin and mmax2 definitions depend on the mass prior model.          
        '''

        self.mmax = self.mmaxbh #Maximum value of m1, used in injections.Injections.update_VT (m_prior.mmax)
        self.mmin = self.mminbh #Minimum value of m2, used in self.joint_prob and in injections.Injections.update_VT (m_prior.mmin) 
        self.mmax2 = self.mmaxbh #Maximum value of m2, used in self.joint_prob

        self.mdis={'mass_1':_cmp.PowerLaw_math(alpha=-self.alpha,min_pl=self.mminbh,max_pl=self.mmaxbh),
                     'mass_2':_cmp.PowerLaw_math(alpha=self.beta,min_pl=self.mminbh,max_pl=self.mmaxbh)}

class NSBH_powerlaw(m_priors):
    """
    Child class for NS-BH power law distribution.
    
    Parameters
    -------------
    mminbh: Minimum mass of the PL component of the black hole mass distribution
    mmaxbh: Maximum mass of the PL component of the black hole mass distribution
    alpha: Spectral index for the PL of the primary mass distribution
    mminns: Minimum mass of the neutron star distribution
    mmaxns: Maximum mass of the neutron star distribution
    alphans: Spectral index for the PL of the neutron star mass distribution

    The default values of the black hole mass distribution parameters are set to the corresponding median values in the uniform priors reported in 2111.03604
    The default values of the neutron star mass distribution parameters are set to the corresponding values reported in section 4.2 (page 23) in 2111.03604

    ************
    NOTE: The spectral indices passed to PowerLaw_math are alpha=-self.alpha, and alpha=-self.alphans, according to eq. A10 in 2111.03604
    *************   

    The method m_priors.update_parameters is used in the constructor to initialize the objects
    """
    def __init__(self,mminbh=6.0,mmaxbh=125.0,alpha=6.75,mminns=1.0,mmaxns=3.0,alphans=0.0):
        super().__init__()

        self.update_parameters(param_dict={'alpha':alpha, 'mminbh':mminbh, 'mmaxbh':mmaxbh, 'alphans':alphans, 'mminns':mminns, 'mmaxns':mmaxns})
        
    def update_mass_priors(self):
        ''' 
        This method creates a dictionary of mass distributions objects.         
        It sets the maximum value of the primary mass distribution mmax to mmaxbh, 
        the minimum value of the secondary mass distribution mmin to mminns, 
        and the maximum value of the secondary mass distribution mmax2 to mmaxns.
        It's called by update_paratemters everytime the mass priors parameters are changed.
        Every mass priors model has a different implementation because the distributions are different,
        and mmax, mmin and mmax2 definitions depend on the mass prior model.          
        '''

        self.mmax=self.mmaxbh
        self.mmin=self.mminns        
        self.mmax2=self.mmaxns

        self.mdis={'mass_1':_cmp.PowerLaw_math(alpha=-self.alpha,min_pl=self.mminbh,max_pl=self.mmaxbh),
                     'mass_2':_cmp.PowerLaw_math(alpha=-self.alphans,min_pl=self.mminns,max_pl=self.mmaxns)}

class BBH_powerlaw_gaussian(m_priors):
    """
    Child class for BBH power law gaussian distribution.
    
    Parameters
    -------------
    mminbh: Minimum mass of the PL component of the black hole mass distribution
    mmaxbh: Maximum mass of the PL component of the black hole mass distribution
    alpha: Spectral index for the PL of the primary mass distribution    
    mu_g: Mean of the Gaussian component in the primary mass distribution
    sigma_g: Width of the Gaussian component in the primary mass distribution
    lambda_peak: Fraction of the model in the Gaussian component
    delta_m: Range of mass tapering on the lower end of the mass distribution
    beta: Spectral index for the PL of the mass ratio distribution

    The default values of the parameters are set to the corresponding values reported in section 4.2 (page 23) in 2111.03604

    ************
    NOTE: The spectral indices passed to PowerLawGaussian_math, and PowerLaw_math, are alpha=-self.alpha, and alpha=self.beta, according to eqs. A8,A11 in 2111.03604
    *************   

    The method m_priors.update_parameters is used in the constructor to initialize the objects.
    """
    def __init__(self,mminbh=4.98,mmaxbh=112.5,alpha=3.78,mu_g=32.27,sigma_g=3.88,lambda_peak=0.03,delta_m=4.8,beta=0.81):
        super().__init__()

        self.update_parameters(param_dict={'alpha':alpha, 'beta':beta, 'mminbh':mminbh, 'mmaxbh':mmaxbh, 'mu_g':mu_g, 'sigma_g':sigma_g, 'lambda_peak':lambda_peak, 'delta_m':delta_m})

    def update_mass_priors(self):
        ''' 
        This method creates a dictionary of mass distributions objects.         
        It sets the maximum value of the primary mass distribution mmax to self.mdis['mass_1'].maximum, 
        the minimum value of the secondary mass distribution mmin to mminbh, 
        and the maximum value of the secondary mass distribution mmax2 to mmaxbh.
        It's called by update_paratemters everytime the mass priors parameters are changed.
        Every mass priors model has a different implementation because the distributions are different,
        and mmax, mmin and mmax2 definitions depend on the mass prior model.          
        '''
                       
        self.m1pr = _cmp.PowerLawGaussian_math(alpha=-self.alpha,min_pl=self.mminbh,max_pl=self.mmaxbh,lambda_g=self.lambda_peak
                    ,mean_g=self.mu_g,sigma_g=self.sigma_g,min_g=self.mminbh,max_g=self.mu_g+5*self.sigma_g)

        # The max of the secondary mass is adapted to the primary mass maximum which is desided byt the Gaussian and PL
        self.m2pr = _cmp.PowerLaw_math(alpha=self.beta,min_pl=self.mminbh,max_pl=np.max([self.mu_g+5*self.sigma_g,self.mmaxbh]))

        self.mdis={'mass_1': _cmp.SmoothedProb(origin_prob=self.m1pr,bottom=self.mminbh,bottom_smooth=self.delta_m),
                      'mass_2':_cmp.SmoothedProb(origin_prob=self.m2pr,bottom=self.mminbh,bottom_smooth=self.delta_m)}
       
        # TO DO Add a check on the mu_g - 5 sigma of the gaussian to not overlap with mmin, print a warning
        #if (mu_g - 5*sigma_g)<=mmin:
        #print('Warning, your mean (minuse 5 sigma) of the gaussian component is too close to the minimum mass')

        self.mmax = self.mdis['mass_1'].maximum 
        self.mmin = self.mminbh  
        self.mmax2 = self.mmaxbh


class NSBH_powerlaw_gaussian(m_priors):
    """
    Child class for NS-BH power law gaussian distribution.
    
    Parameters
    -------------
    mminbh: Minimum mass of the PL component of the black hole mass distribution
    mmaxbh: Maximum mass of the PL component of the black hole mass distribution
    alpha: Spectral index for the PL of the primary mass distribution    
    mu_g: Mean of the Gaussian component in the primary mass distribution
    sigma_g: Width of the Gaussian component in the primary mass distribution
    lambda_peak: Fraction of the model in the Gaussian component    
    delta_m: Range of mass tapering on the lower end of the mass distribution
    mminns: Minimum mass of the neutron star distribution
    mmaxns: Maximum mass of the neutron star distribution
    alphans: Spectral index for the PL of the neutron star mass distribution

    The default values of the parameters are set to the corresponding values reported in section 4.2 (page 23) in 2111.03604

    ************
    NOTE: The spectral indices passed to PowerLawGaussian_math, and PowerLaw_math, are alpha=-self.alpha, and alpha=-self.alphans, according to eqs. A10,A11 in 2111.03604
    *************
        
    The method m_priors.update_parameters is used in the constructor to initialize the objects.
    """
    def __init__(self,mminbh=4.98,mmaxbh=112.5,alpha=3.78,mu_g=32.27,sigma_g=3.88,lambda_peak=0.03,delta_m=4.8,mminns=1.0,mmaxns=3.0,alphans=0.0):
        super().__init__()

        self.update_parameters(param_dict={'alpha':alpha, 'mminbh':mminbh, 'mmaxbh':mmaxbh, 'mu_g':mu_g, 'sigma_g':sigma_g, 'lambda_peak':lambda_peak, 'delta_m':delta_m, 'alphans':alphans, 'mminns':mminns, 'mmaxns':mmaxns})

    def update_mass_priors(self):
        ''' 
        This method creates a dictionary of mass distributions objects.         
        It sets the maximum value of the primary mass distribution mmax to self.mdis['mass_1'].maximum, 
        the minimum value of the secondary mass distribution mmin to mminns, 
        and the maximum value of the secondary mass distribution mmax2 to mmaxns.
        It's called by update_paratemters everytime the mass priors parameters are changed.
        Every mass priors model has a different implementation because the distributions are different,
        and mmax, mmin and mmax2 definitions depend on the mass prior model.          
        '''
                
        self.m1pr = _cmp.PowerLawGaussian_math(alpha=-self.alpha,min_pl=self.mminbh,max_pl=self.mmaxbh,lambda_g=self.lambda_peak
                    ,mean_g=self.mu_g,sigma_g=self.sigma_g,min_g=self.mminbh,max_g=self.mu_g+5*self.sigma_g)

        # The max of the secondary mass is adapted to the primary mass maximum which is desided byt the Gaussian and PL
        self.m2pr = _cmp.PowerLaw_math(alpha=-self.alphans,min_pl=self.mminns,max_pl=self.mmaxns)

        self.mdis={'mass_1': _cmp.SmoothedProb(origin_prob=self.m1pr,bottom=self.mminbh,bottom_smooth=self.delta_m),
                      'mass_2':self.m2pr}

        self.mmax = self.mdis['mass_1'].maximum 
        self.mmin = self.mminns  
        self.mmax2 = self.mmaxns

class BBH_broken_powerlaw(m_priors):
    """
    Child class for BBH broken power law distribution.

    Parameters
    -------------
    mminbh: Minimum mass of the PL component of the black hole mass distribution
    mmaxbh: Maximum mass of the PL component of the black hole mass distribution
    alpha_1: PL slope of the primary mass distribution for masses below mbreak 
    alpha_2: PL slope for the primary mass distribution for masses above mbreak 
    b: The fraction of the way between mminbh and mmaxbh at which the primary mass distribution breaks
    delta_m: Range of mass tapering on the lower end of the mass distribution
    beta: Spectral index for the PL of the mass ratio distribution

    The default values of the parameters are set to the corresponding median values in the uniform priors reported in 2111.03604
    
    ************
    NOTE: The spectral indices passed to BrokenPowerLaw_math, and PowerLaw_math, are alpha_1=-self.alpha_1, alpha_2=-self.alpha_2, and alpha=self.beta, according to eqs. A8,A12 in 2111.03604
    ************

    The method m_priors.update_parameters is used in the constructor to initialize the objects.
    """
    def __init__(self,mminbh=26,mmaxbh=125,alpha_1=6.75,alpha_2=6.75,b=0.5,delta_m=5,beta=4):
        super().__init__()

        self.update_parameters(param_dict={'alpha_1':alpha_1, 'alpha_2':alpha_2, 'beta':beta, 'mminbh':mminbh, 'mmaxbh':mmaxbh, 'b':b, 'delta_m':delta_m})

    def update_mass_priors(self):
        ''' 
        This method creates a dictionary of mass distributions objects. 
        It sets the maximum value of the primary mass distribution mmax to mmaxbh, 
        the minimum value of the secondary mass distribution mmin to mminbh, 
        and the maximum value of the secondary mass distribution mmax2 to mmaxbh.
        It's called by update_paratemters everytime the mass priors parameters are changed.
        Every mass priors model has a different implementation because the distributions are different,
        and mmax, mmin and mmax2 definitions depend on the mass prior model.          
        '''

        self.mmax = self.mmaxbh 
        self.mmin = self.mminbh  
        self.mmax2 = self.mmaxbh
                
        self.m1pr = _cmp.BrokenPowerLaw_math(alpha_1=-self.alpha_1,alpha_2=-self.alpha_2,min_pl=self.mminbh,max_pl=self.mmaxbh,b=self.b)
        self.m2pr = _cmp.PowerLaw_math(alpha=self.beta,min_pl=self.mminbh,max_pl=self.mmaxbh)

        self.mdis={'mass_1': _cmp.SmoothedProb(origin_prob=self.m1pr,bottom=self.mminbh,bottom_smooth=self.delta_m),
                      'mass_2':_cmp.SmoothedProb(origin_prob=self.m2pr,bottom=self.mminbh,bottom_smooth=self.delta_m)}

class NSBH_broken_powerlaw(m_priors):
    """
    Child class for NS-BH broken power law distribution.
    
    Parameters
    -------------
    mminbh: Minimum mass of the PL component of the black hole mass distribution
    mmaxbh: Maximum mass of the black hole mass distribution
    alpha_1: PL slope of the primary mass distribution for masses below mbreak 
    alpha_2: PL slope for the primary mass distribution for masses above mbreak 
    b: The fraction of the way between mminbh and mmaxbh at which the primary mass distribution breaks
    delta_m: Range of mass tapering on the lower end of the mass distribution
    mminns: Minimum mass of the neutron star distribution
    mmaxns: Maximum mass of the neutron star distribution
    alphans: Spectral index for the PL of the neutron star mass distribution

    The default values of the black hole mass distribution parameters are set to the corresponding median values in the uniform priors reported in 2111.03604
    The default values of the neutron star mass distribution parameters are set to the corresponding values reported in section 4.2 (page 23) in 2111.03604

    ************
    NOTE: The spectral indices passed to BrokenPowerLaw_math, and PowerLaw_math, are alpha_1=-self.alpha_1, alpha_2=-self.alpha_2, and alpha=-self.alphans, according to eqs. A10,A12 in 2111.03604
    ************
    
    The method m_priors.update_parameters is used in the constructor to initialize the objects.
    """
    def __init__(self,mminbh=26,mmaxbh=125,alpha_1=6.75,alpha_2=6.75,b=0.5,delta_m=5,mminns=1.0,mmaxns=3.0,alphans=0.0):
        super().__init__()

        self.update_parameters(param_dict={'alpha_1':alpha_1, 'alpha_2':alpha_2, 'mminbh':mminbh, 'mmaxbh':mmaxbh, 'b':b, 'delta_m':delta_m, 'alphans':alphans, 'mminns':mminns, 'mmaxns':mmaxns})

    def update_mass_priors(self):
        ''' 
        This method creates a dictionary of mass distributions objects.         
        It sets the maximum value of the primary mass distribution mmax to mmaxbh, 
        the minimum value of the secondary mass distribution mmin to mminns, 
        and the maximum value of the secondary mass distribution mmax2 to mmaxns.
        It's called by update_paratemters everytime the mass priors parameters are changed.
        Every mass priors model has a different implementation because the distributions are different,
        and mmax, mmin and mmax2 definitions depend on the mass prior model.          
        '''
        
        self.mmax=self.mmaxbh
        self.mmin=self.mminns        
        self.mmax2=self.mmaxns
                
        self.m1pr = _cmp.BrokenPowerLaw_math(alpha_1=-self.alpha_1,alpha_2=-self.alpha_2,min_pl=self.mminbh,max_pl=self.mmaxbh,b=self.b)
        self.m2pr = _cmp.PowerLaw_math(alpha=-self.alphans,min_pl=self.mminns,max_pl=self.mmaxns)

        self.mdis={'mass_1': _cmp.SmoothedProb(origin_prob=self.m1pr,bottom=self.mminbh,bottom_smooth=self.delta_m),
                      'mass_2':self.m2pr}

class BNS(m_priors):
    """
    Child class for BNS distribution.
    
    Parameters
    -----------
    mminns: Minimum mass of the neutron star distribution
    mmaxns: Maximum mass of the neutron star distribution
    alphans: Spectral index for the PL of the neutron star mass distribution

    The default values of the parameters are set to the corresponding values reported in section 4.2 (page 23) in 2111.03604

    ************
    NOTE: The spectral index passed to PowerLaw_math is alpha=-self.alphans according to eq. A10 in 2111.03604
    ************
    
    The method m_priors.update_parameters is used in the constructor to initialize the objects.
    """
    def __init__(self,mminns=1.0,mmaxns=3.0,alphans=0.0):
        super().__init__()

        self.update_parameters(param_dict={'alphans':alphans, 'mminns':mminns, 'mmaxns':mmaxns})
                    
    def update_mass_priors(self):
        ''' 
        This method creates a dictionary of mass distributions objects.         
        It sets the maximum value of the primary mass distribution mmax to mmaxns, 
        and the minimum value of the secondary mass distribution mmin to mminns.
        It's called by update_paratemters everytime the mass priors parameters are changed.
        Every mass priors model has a different implementation because the distributions are different,
        and mmax and mmin definitions depend on the mass prior model.          
        '''

        self.mmax = self.mmaxns
        self.mmin = self.mminns        

        self.mdis={'mass_1':_cmp.PowerLaw_math(alpha=-self.alphans,min_pl=self.mminns,max_pl=self.mmaxns),
                  'mass_2':_cmp.PowerLaw_math(alpha=-self.alphans,min_pl=self.mminns,max_pl=self.mmaxns)}
        
    
    def joint_prob(self, ms1, ms2):
        """ 
        This method returns the joint probability :math:`p(m_1,m_2)`

        Parameters
        ----------
        ms1: np.array(matrix)
            mass one in solar masses
        ms2: dict
            mass two in solar masses
        """

        to_ret =self.mdis['mass_1'].prob(ms1)*self.mdis['mass_2'].prob(ms2)

        return to_ret
    

    def sample(self, Nsample):
        """
        This method samples from the joint probability :math:`p(m_1,m_2)`

        Parameters
        ----------
        Nsample: int
            Number of samples you want
        """

        vals_m1 = np.random.rand(Nsample)
        vals_m2 = np.random.rand(Nsample)

        m1_trials = np.logspace(np.log10(self.mdis['mass_1'].minimum),np.log10(self.mdis['mass_1'].maximum),10000)
        m2_trials = np.logspace(np.log10(self.mdis['mass_2'].minimum),np.log10(self.mdis['mass_2'].maximum),10000)

        cdf_m1_trials = self.mdis['mass_1'].cdf(m1_trials)
        cdf_m2_trials = self.mdis['mass_2'].cdf(m2_trials)

        m1_trials = np.log10(m1_trials)
        m2_trials = np.log10(m2_trials)

        _,indxm1 = np.unique(cdf_m1_trials,return_index=True)
        _,indxm2 = np.unique(cdf_m2_trials,return_index=True)

        interpo_icdf_m1 = interp1d(cdf_m1_trials[indxm1],m1_trials[indxm1],bounds_error=False,fill_value=(m1_trials[0],m1_trials[-1]))
        interpo_icdf_m2 = interp1d(cdf_m2_trials[indxm2],m2_trials[indxm2],bounds_error=False,fill_value=(m2_trials[0],m2_trials[-1]))

        mass_1_samples = 10**interpo_icdf_m1(vals_m1)
        mass_2_samples = 10**interpo_icdf_m2(vals_m2)

        indx = np.where(mass_2_samples>mass_1_samples)[0]
        mass_1_samples[indx],mass_2_samples[indx] = mass_2_samples[indx],mass_1_samples[indx]

        return mass_1_samples, mass_2_samples



"""
Checking the valid interpolation boundary of parameters given the injections.

Anson Chen
"""

import numpy as np
import copy
import logging


def check_boundary(cosmo, parameter_dict, injections, mass_priors, gravity_model, mass_model):
    """
    WARNING, CURRENTLY UNREVIEWED
    """
    
    logging.warning("This part of gwcosmo is unreviewed: proceed with caution.")

    injections = copy.deepcopy(injections)

    if mass_model == 'BNS':
        mass_priors.mmax = parameter_dict['mmaxns']['value']
    else:
        mass_priors.mmax = parameter_dict['mmaxbh']['value']

    if np.size(parameter_dict['H0']['value'])>1:
        H0_max = parameter_dict['H0']['value'][1]
    else:
        H0_max = parameter_dict['H0']['value']

    if gravity_model == 'GR':

        # highest redshift case
        cosmo.H0 = H0_max

        if max(injections.dldet*H0_max/cosmo.c) > max(cosmo.dlH0overc_z_arr):
            H0_bound = cosmo.c*max(cosmo.dlH0overc_z_arr)/max(injections.dldet)
            raise ValueError('Injection GW distances exceed the LOS redshift prior bound. The maximum H0 higher bound is %.1f.' % H0_bound)

        z_inj_max = cosmo.z_dgw(max(injections.dldet))
        Msmax = max(injections.m1det)/(1+z_inj_max)

        if Msmax < mass_priors.mmax:
            raise ValueError('For the extreme value of H0=%.1f, the source-frame prior becomes inconsistent with the injections at a dL of %.1f Mpc, which is within the detection horizon for the current set up. Please consider a lower M_max for mass prior (e.g. %.1f Msol), or a lower H0_max.'%(H0_max,max(injections.dldet),Msmax+0.1))
        else:
            print('Source-frame prior M_max=%.1f Msol is consistent with the injections at a dL of %.1f Mpc at H0_max=%.1f'%(mass_priors.mmax,max(injections.dldet),H0_max))

    elif gravity_model == 'Xi0_n':
        if np.size(parameter_dict['Xi0']['value'])>1:
            Xi0_min = parameter_dict['Xi0']['value'][0]
        else:
            Xi0_min = parameter_dict['Xi0']['value']
        if np.size(parameter_dict['n']['value'])>1:
            n_max = parameter_dict['n']['value'][1]
        else:
            n_max = parameter_dict['n']['value']

        # highest redshift case
        cosmo_param_dict = {'H0': H0_max, 'Xi0': Xi0_min, 'n': n_max}
        cosmo.update_parameters(cosmo_param_dict)

        if max(injections.dldet*H0_max/cosmo.c) > cosmo.dgw_dL_ratio(cosmo.zmax)*max(cosmo.dlH0overc_z_arr):
            z_ratio = 1/(1+cosmo.zmax)**n_max
            Xi0_bound = (max(injections.dldet*H0_max/cosmo.c)/max(cosmo.dlH0overc_z_arr)-z_ratio) / (1-z_ratio)
            raise ValueError('Injection GW distances exceed the redshift prior bound. For the H0 upper bound of %d and the n upper bound of %.2f, the minimum Xi0 lower bound is %.2f.' %(H0_max, n_max, Xi0_bound+0.01))

        z_inj_max = cosmo.z_dgw(max(injections.dldet))
        Msmax = max(injections.m1det)/(1+z_inj_max)

        if Msmax < mass_priors.mmax:
            raise ValueError('For the extreme values of H0=%.1f, Xi0=%.2f, n=%.2f, the source-frame prior becomes inconsistent with the injections at a dL of %.1f Mpc, which is within the detection horizon for the current set up. Please consider a lower M_max for mass prior (e.g. %.1f Msol), a lower H0_max or a higher Xi0_min.'%(H0_max,Xi0_min,n_max,max(injections.dldet),Msmax+0.1))
        else:
            print('Source-frame prior M_max=%.1f Msol is consistent with the injections at a dL of %.1f Mpc at H0_max=%.1f, Xi0_min=%.2f, n_max=%.2f'%(mass_priors.mmax,max(injections.dldet),H0_max,Xi0_min,n_max))




import gwcosmo
import numpy as np
import matplotlib.pyplot as plt
import healpy as hp
from scipy.integrate import quad
from scipy.stats import norm
from scipy.interpolate import interp1d

from gwcosmo.utilities import schechter_params
from gwcosmo.utilities import cosmology
from gwcosmo.utilities import luminosity_function, calc_kcor
from gwcosmo.prior.catalog import color_names

from ligo.skymap.io import fits
from ligo.skymap import plot
from ligo.skymap import postprocess

from tqdm import tqdm


def mth2hpx(ra, dec, m, nside):
    """
    Map an array of values ``m'' defined for the sky points ra and dec in a
    HEALPY array.

    Parameters
    ----------
    ra, dec : (ndarray, ndarray)
        Coordinates of the sources in radians.

    nside : int
        HEALPix nside of the target map

    Return
    ------
    hpx_map : ndarray
        HEALPix map of the catalogue map.

    """

    # The number of pixels based on the chosen value of nside
    npix = hp.nside2npix(nside)

    # conver to theta, phi
    theta = np.pi/2.0 - dec
    phi = ra

    # convert to HEALPix indices (each galaxy is assigned to a single healpy pixel)
    indices = hp.ang2pix(nside, theta, phi)

    # sort the indices into ascending order
    idx_sort = np.argsort(indices)
    sorted_indices = indices[idx_sort]

    # idx: the healpy index of each pixel containing a galaxy (arranged in ascending order)
    # idx_start: the index of 'sorted_indices' corresponding to each new pixel
    # count: the number of galaxies in each pixel
    idx, idx_start,count = np.unique(sorted_indices,return_counts=True,return_index=True)

    # splits indices into arrays - 1 per pixel
    res = np.split(idx_sort, idx_start[1:])

    mths = np.zeros(npix)

    marray = np.array(m)
    count=0
    for i in range(npix):
        if i in idx:
            # select all the apparent magnitudes of galaxies in the ith pixel
            ms = marray[res[count]]
            count += 1
            # find the median apparent magnitude within this pixel (mth stays as zero if no galaxies are present)
            mths[i] = np.median(ms)

    # fill the fullsky map
    hpx_map = np.zeros(npix)
    hpx_map[range(npix)] = mths

    return hpx_map

def overdensity_given_skypos(galaxy_cat,band,Schparam,ra,dec,zproxy,nside=64,h0=67.7
                             ,cosmo=cosmology.standard_cosmology(Omega_m=0.308, w0=-1., wa=0., zmax=10.0),lweight=True,redunc=False):
    """
    This function returns the redshift prior from the galaxy catalog, from the empty-catalog case and the
    index corresponsing to galaxies in the pixel corresponding to a given sky location.

    Parameters
    ----------
    galaxy_cat: galaxy catalog from gwcosmo
    Schparam: str
        Schechter class for the parameters
    ra: float
        right ascension in radiants where to compute the priors
    dec: float
        declination in radiants where to compute the priors
    zproxy: numpy array
        Redshift at which to calculate the overdensity
    nside: integer
        Nside parameter for the sky division in pixels by HEALPY, must be power of 2
    h0: float
        Hubble constant
    cosmo: gwcosmo standard cosmology
        GWcosmo standard cosmology class
    lweight: bool
        If to use luminosity weightening or not
    redunc: bool
        False if you dont want redshift uncertainties. Note that using redshift uncertainties will only generate a proxy figure (true distribution)
        convoluted with error function.
    """

    cosmo.H0 = h0
    to_return = np.zeros_like(zproxy)
    # The number of pixels based on the chosen value of nside
    npix = hp.nside2npix(nside)
    theta = np.pi/2.0 - galaxy_cat.data['dec']
    phi = galaxy_cat.data['ra']
    indices = hp.ang2pix(nside, theta, phi)
    index_point = hp.ang2pix(nside, np.pi/2.0 - dec, ra)
    gal_index = np.where(indices==index_point)[0]

    if not gal_index.size:
        print('No galaxy overlapping with ra={:.2f} and dec={:.2f}'.format(ra,dec))
        return to_return

    allm=galaxy_cat.data['m_'+band][gal_index]
    allra=galaxy_cat.data['ra'][gal_index]
    alldec=galaxy_cat.data['dec'][gal_index]
    allz=galaxy_cat.data['z'][gal_index]
    allsigmaz=galaxy_cat.data['sigmaz'][gal_index]

    mth = np.median(allm)
    selected_galaxies = np.where(allm<mth)[0]

    allm=allm[selected_galaxies]
    allra=allra[selected_galaxies]
    alldec=alldec[selected_galaxies]
    allz=allz[selected_galaxies]
    allsigmaz=allsigmaz[selected_galaxies]
    gal_index=gal_index[selected_galaxies]

    Kcorr = galaxy_cat.get_k_correction(band,allz,color_names[band],color_value=allm)
    lweights = gwcosmo.utilities.luminosity_function.L_mdl(allm, cosmo.dl_zH0(allz),Kcorr=Kcorr)

    alpha,Mstar_obs,Mmin_source,Mmax_source = Schparam.alpha,Schparam.Mstar,Schparam.Mmin,Schparam.Mmax

    Lmax=gwcosmo.utilities.luminosity_function.L_M(Mmin_source)
    Lmin=gwcosmo.utilities.luminosity_function.L_M(Mmax_source)

    index_lum = np.where((lweights<Lmax) & (lweights>Lmin))[0]
    allm=allm[index_lum]
    allra=allra[index_lum]
    alldec=alldec[index_lum]
    allz=allz[index_lum]
    allsigmaz=allsigmaz[index_lum]
    lweights=lweights[index_lum]
    gal_index=gal_index[index_lum]
    zprior = cosmo.p_z
    
    if redunc:
        for i in range(len(allz)):
            zcut = np.linspace(0,allz[i]+9*allsigmaz[i],5000)
            priordens=zprior(zcut)
            zl = (1./(1+zcut))*priordens*norm.pdf(zcut,allz[i],allsigmaz[i])
            zl /=np.trapz(zl,zcut)
            interpo = interp1d(zcut,zl,bounds_error=False,fill_value=0.)

            if lweight:
                to_return+=interpo(zproxy)*lweights[i]*skymap_gwcosmo.skyprob(allra[i],alldec[i])
            else:
                to_return+=interpo(zproxy)*skymap_gwcosmo.skyprob(allra[i],alldec[i])
        galdens= to_return/np.trapz(to_return,zproxy)
    else:
        edges=zproxy-0.5*(zproxy[1]-zproxy[0])
        edges=np.append(edges,zproxy[-1]+0.5*(zproxy[1]-zproxy[0]))
        
        if lweight:
            to_return, edged = np.histogram(allz,bins=edges,weights=lweights/(1+allz))
        else:
            to_return, edged = np.histogram(allz,bins=edges,weights=1/(1+allz))
        galdens= to_return/np.sum(to_return*(zproxy[1]-zproxy[0]))    

    priordens=zprior(zproxy)
    priordens/=np.trapz(priordens,zproxy)
    return galdens,priordens,gal_index


def overdensity_given_GWskyarea(galaxy_cat,band,Schparam,fits_file,CL,zproxy,h0=67.7,nside=None
                             ,cosmo=cosmology.standard_cosmology(Omega_m=0.308, w0=-1., wa=0., zmax=10.0),lweight=True,redunc=False):
    """
    This function returns the redshift prior from the galaxy catalog, from the empty-catalog case and the
    index corresponsing to galaxies in the pixel corresponding to a given sky location.

    Parameters
    ----------
    galaxy_cat: galaxy catalog from gwcosmo
    Schparam: str
        Schechter class for the parameters
    fits_file: string
        path to the GWskymap
    CL: float
        What confidence level interval you want,e.g. 95
    zproxy: numpy array
        Redshift at which to calculate the overdensity
    h0: float
        Hubble constant
    cosmo: gwcosmo standard cosmology
        GWcosmo standard cosmology class
    lweight: bool
        If to use luminosity weightening or not
    redunc: bool
        False if you dont want redshift uncertainties. Note that using redshift uncertainties will only generate a proxy figure (true distribution)
        convoluted with error function.
    """

    cosmo.H0 = h0

    skymap_gwcosmo = gwcosmo.likelihood.skymap.skymap(fits_file)
    gal_index = np.where(skymap_gwcosmo.samples_within_region(galaxy_cat.data['ra'],galaxy_cat.data['dec'],CL,nside=nside))[0]
    to_return = np.zeros_like(zproxy)

    if not gal_index.size:
        print('No galaxy overlapping')
        return to_return

    allm=galaxy_cat.data['m_'+band][gal_index]
    allra=galaxy_cat.data['ra'][gal_index]
    alldec=galaxy_cat.data['dec'][gal_index]
    allz=galaxy_cat.data['z'][gal_index]
    allsigmaz=galaxy_cat.data['sigmaz'][gal_index]

    mth = np.median(allm)
    selected_galaxies = np.where(allm<mth)[0]

    allm=allm[selected_galaxies]
    allra=allra[selected_galaxies]
    alldec=alldec[selected_galaxies]
    allz=allz[selected_galaxies]
    allsigmaz=allsigmaz[selected_galaxies]
    gal_index=gal_index[selected_galaxies]


    Kcorr = galaxy_cat.get_k_correction(band,allz,color_names[band],color_value=allm)
    lweights = gwcosmo.utilities.luminosity_function.L_mdl(allm, cosmo.dl_zH0(allz),Kcorr=Kcorr)

    alpha,Mstar_obs,Mmin_source,Mmax_source = Schparam.alpha,Schparam.Mstar,Schparam.Mmin,Schparam.Mmax

    Lmax=gwcosmo.utilities.luminosity_function.L_M(Mmin_source)
    Lmin=gwcosmo.utilities.luminosity_function.L_M(Mmax_source)

    index_lum = np.where((lweights<Lmax) & (lweights>Lmin))[0]
    allm=allm[index_lum]
    allra=allra[index_lum]
    alldec=alldec[index_lum]
    allz=allz[index_lum]
    allsigmaz=allsigmaz[index_lum]
    lweights=lweights[index_lum]
    gal_index=gal_index[index_lum]
    zprior = cosmo.p_z
    
    if redunc:
        for i in range(len(allz)):
            zcut = np.linspace(0,allz[i]+9*allsigmaz[i],5000)
            priordens=zprior(zcut)
            zl = (1./(1+zcut))*priordens*norm.pdf(zcut,allz[i],allsigmaz[i])
            zl /=np.trapz(zl,zcut)
            interpo = interp1d(zcut,zl,bounds_error=False,fill_value=0.)

            if lweight:
                to_return+=interpo(zproxy)*lweights[i]*skymap_gwcosmo.skyprob(allra[i],alldec[i])
            else:
                to_return+=interpo(zproxy)*skymap_gwcosmo.skyprob(allra[i],alldec[i])
        galdens= to_return/np.trapz(to_return,zproxy)
    else:
        edges=zproxy-0.5*(zproxy[1]-zproxy[0])
        edges=np.append(edges,zproxy[-1]+0.5*(zproxy[1]-zproxy[0]))
        
        if lweight:
            to_return, edged = np.histogram(allz,bins=edges,weights=lweights/(1+allz))
        else:
            to_return, edged = np.histogram(allz,bins=edges,weights=1/(1+allz))
        galdens= to_return/np.sum(to_return*(zproxy[1]-zproxy[0]))    

    priordens=zprior(zproxy)
    priordens/=np.trapz(priordens,zproxy)
    return galdens,priordens,gal_index


def Completeness(H0,z_array,mth,Schparam,weighted=True):
    """
    Returns p(G|H0,z)
    The probability that the host galaxy is in the catalogue given detection and H0.

    Parameters
    ----------
    H0 : float
        Hubble constant value in kms-1Mpc-1
    z_array : float, array_like
        An array of redshifts
    mth : float
        The apparent magnitude threshold of the galaxy catalog under consideration
    Schparam: str
        Schechter class for the parameters
    pdet_path: str
        path to pdet
    weighted : bool, optional
        Is the GW host luminosity weighted? (default=False)

    Returns
    -------
    float or array_like
        p(G|H0,z)
    """

    alpha,Mstar_obs,Mmin_source,Mmax_source = Schparam.alpha,Schparam.Mstar,Schparam.Mmin,Schparam.Mmax

    cosmo = cosmology.standard_cosmology(H0=H0)
    zprior = cosmo.p_z
    num = np.zeros(len(z_array))
    den = np.zeros(len(z_array))

    for i in range(len(z_array)):

        def I(M):
            temp = luminosity_function.SchechterMagFunction(alpha=alpha,Mstar_obs=Mstar_obs)(M,H0)
            if weighted:
                return temp*luminosity_function.L_M(M)
            else:
                return temp

        Mmin = luminosity_function.M_Mobs(H0,Mmin_source)
        Mmax = luminosity_function.M_Mobs(H0,Mmax_source)
        lim = luminosity_function.M_mdl(mth,cosmo.dl_zH0(z_array[i]))
        if lim > Mmax:
            num[i]=1.0
            den[i]=1.0
        else:
            num[i] = quad(I,Mmin,lim,epsabs=1.49e-13,epsrel=1.49e-13)[0]
            den[i] = quad(I,Mmin,Mmax,epsabs=1.49e-13,epsrel=1.49e-13)[0]

    pGD = num/den
    return pGD


def plot_skymap(map_plot,mask=None,ax=None,**kwargs):
    """
    Plot a skymap using contourf of an HEALPY array:

    Parameters
    ----------
    map_plot: healpy array
        An healpy array with the values of the map you want to plot inside
    mask: array
        Array of the same size of the numpy array with False and True.
    ax: matplolib lib axes
        Axes where to plot.
    **kwargs:
        Keyword argumets to pass to contourf

    Returns
    -------
    axes handling, image handling for skymaps
    """

    if ax is None:
        ax = plt.axes(projection='astro hours mollweide')

    if mask is not None:
        indx0= np.where(mask)[0]

    im=ax.contourf_hpx(map_plot,**kwargs)
    return ax,im

def plot_loc_contour(fits_file,ax=None,**kwargs):
    """
    Plot the CL counters of a skymap

    Parameters
    ----------
    fits_file: fits file for the GW skymap
    ax: matplolib lib axes
        Axes where to plot.
    **kwargs:
        Keyword argumets to pass to contour. If you want to plot levels of CL, provide
        levels=[68,90] etc.
    Returns
    -------
    axes handling, image handling for skymaps
    """

    if ax is None:
        ax = plt.axes(projection='astro hours mollweide')
    skymap, metadata = fits.read_sky_map(fits_file, nest=None)
    cls = 100 * postprocess.find_greedy_credible_levels(skymap)
    cs = ax.contour_hpx((cls, 'ICRS'), nested=metadata['nest'],**kwargs)
    return ax, cs



import os
def get_cachedir(cachedir=None):
    if cachedir is None:
        if 'GWCOSMO_CACHE' in os.environ.keys():
            cachedir = os.environ['GWCOSMO_CACHE']
        else:
            cachedir = os.path.join(os.environ['HOME'], '.cache/gwcosmo')
    os.makedirs(cachedir, exist_ok=True)
    return cachedir



"""
Module with Schechter magnitude function:
(C) Walter Del Pozzo (2014)
Rachel Gray
"""
from numpy import *
from scipy.integrate import quad

class SchechterMagFunction(object):
    """
    The Schechter function (as a function of absolute magnitude and H0)
    """
    
    def __init__(self, Mstar_obs=-19.70, alpha=-1.07, phistar=1.):
        """
        Parameters
        ----------
        Mstar_obs : float, optional
            Observed absolute magnitude, M* - 5log10(h) (default=-19.7)
        alpha : float, optional
            Faint-end slope parameter (default=-1.07)
        phistar : float, optional
            Normalisation factor defining density of galaxies, Mpc^-3 
            This factor cancels in numerator and denominator of the main 
            analysis, including any H0-dependence (default=1.)
        """
        
        self.Mstar_obs = Mstar_obs
        self.phistar = phistar
        self.alpha = alpha
        self.norm = None

    def evaluate(self, m, H0):
        """
        Evaluate the Schechter function for a given absolute magnitude and H0

        phi(m)=0.4*ln10*phistar*[10^(0.4(Mstar-m)*(alpha+1))]*exp[-10^(0.4(Mstar-m))]
        
        Parameters
        ----------
        m : float
            Absolute magnitude
        H0 : float
            Hubble constant value in kms-1Mpc-1
        
        Returns
        -------
        float
            phi(m)
        """
        
        Mstar = M_Mobs(H0, self.Mstar_obs)
        return 0.4*log(10.0)*self.phistar \
               * pow(10.0, -0.4*(self.alpha+1.0)*(m-Mstar)) \
               * exp(-pow(10, -0.4*(m-Mstar)))

    def normalise(self, mmin, mmax):
        if self.norm is None:
            self.norm = quad(self.evaluate, mmin, mmax)[0]

    def pdf(self, m):
        return self.evaluate(m)/self.norm
        
    def __call__(self, m, H0):
        return self.evaluate(m,H0)


def M_Mobs(H0, M_obs):
    """
    Given an observed absolute magnitude, returns absolute magnitude
    """
    return M_obs + 5.*log10(H0/100.)


# Distance modulus given luminosity distance
def DistanceModulus(dL):
    """
    Returns distance modulus given luminosity distance

    Parameters
    ----------
    dL : luminosity distance in Mpc

    Returns
    -------
    distance modulus = 5*np.log10(dL)+25
    """
    return 5*log10(dL)+25  # dL has to be in Mpc


def dl_mM(m, M, Kcorr=0.):
    """
    returns luminosity distance in Mpc given
    apparent magnitude and absolute magnitude

    Parameters
    ----------
    m : apparent magnitude
    M : absolute magnitude in the source frame
    Kcorr : (optional) K correction, to convert absolute magnitude from the
        observed band to the source frame band (default=0).  If fluxes are
        bolometric, this should be left as 0. If not, a K correction of 0 is
        only valid at low redshifts.

    Returns
    -------
    Luminosity distance in Mpc
    """
    return 10**(0.2*(m-M-Kcorr-25))


def L_M(M):
    """
    Returns luminosity when given an absolute magnitude.
    The constant used here corresponds to the conversion between bolometric mangitude and luminosity.
    It does not matter for the H0 inference, so please use with care when using with band specific magnitudes.

    Parameters
    ----------
    M : absolute magnitude in the source frame

    Returns
    -------
    Luminosity in Watts
    """
    # TODO: double check use of L0=3.0128e28
    return 3.0128e28*10**(-0.4*M)


def M_mdl(m, dl, Kcorr=0.):
    """
    Returns a source's absolute magnitude given
    apparent magnitude and luminosity distance
    If a K correction is supplied it will be applied

    Parameters
    ----------
    m : apparent magnitude
    dl : luminosity distance in Mpc
    Kcorr : (optional) K correction, to convert absolute magnitude from the
        observed band to the source frame band (default=0).  If fluxes are
        bolometric, this should be left as 0. If not, a K correction of 0 is
        only valid at low redshifts.

    Returns
    -------
    Absolute magnitude in the source frame
    """
    return m - DistanceModulus(dl) - Kcorr


def L_mdl(m, dl, Kcorr=0.):
    """
    Returns luminosity when given apparent magnitude and luminosity distance
    If a K correction is supplied it will be applied

    Parameters
    ----------
    m : apparent magnitude
    dl : luminosity distance in Mpc
    Kcorr : (optional) K correction, to convert absolute magnitude from the
        observed band to the source frame band (default=0).  If fluxes are
        bolometric, this should be left as 0. If not, a K correction of 0 is
        only valid at low redshifts.

    Returns
    -------
    Luminosity in the source frame
    """
    return L_M(M_mdl(m, dl, Kcorr=Kcorr))





"""Top-level package for gwcosmo."""
from __future__ import absolute_import
from .utilities import cosmology, luminosity_function, posterior_utilities, redshift_utilities, cache
from .likelihood import posterior_samples, skymap
from .prior import catalog, priors
from .plotting import plot
from .maps import create_mth_map, create_norm_map

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("gwcosmo")
except PackageNotFoundError:
    # package is not installed
    pass



"""
Multi-event likelihood Module
Rachel Gray

"""

import numpy as np

from scipy.integrate import simpson
from scipy.interpolate import interp1d
from gwcosmo.utilities.zprior_utilities import get_zprior_full_sky, get_zprior
from gwcosmo.likelihood.posterior_samples import *
import gwcosmo

from .skymap import ra_dec_from_ipix,ipix_from_ra_dec
import healpy as hp
import bilby
import h5py
import json
import ast
import sys
# import pickle
import copy

class PixelatedGalaxyCatalogMultipleEventLikelihood(bilby.Likelihood):
    """
    Class for preparing and carrying out the computation of the likelihood on 
    H0 for a single GW event
    """
    def __init__(self, posterior_samples_dictionary, skymap_dictionary, injections, LOS_catalog_path, zrates, cosmo, mass_priors, min_pixels=30, sky_area=0.999, posterior_samples_field=None, network_snr_threshold=11.):

        """
        Parameters
        ----------
        samples : object
            GW samples
        skymap : gwcosmo.likelihood.skymap.skymap object
            provides p(x|Omega) and skymap properties
        LOS_prior :
        """
        super().__init__(parameters={'H0': None, 'gamma':None, 'Madau_k':None, 'Madau_zp':None, 'alpha':None, 'delta_m':None, 'mu_g':None, 'sigma_g':None, 'lambda_peak':None, 'alpha_1':None, 'alpha_2':None, 'b':None, 'mminbh':None, 'mmaxbh':None, 'alphans':None, 'mminns':None, 'mmaxns':None, 'beta':None, 'Xi0':None, 'n':None})

        self.zrates = zrates
        
        #TODO make min_pixels an optional dictionary
        LOS_catalog = h5py.File(LOS_catalog_path, 'r')
        temp = LOS_catalog.attrs['opts']
        catalog_opts = ast.literal_eval(temp.decode('utf-8'))
        nside = catalog_opts['nside']
        print(f'Chosen resolution nside: {nside}')
        self.z_array = LOS_catalog['z_array'][:]
        self.zprior_full_sky = get_zprior_full_sky(LOS_catalog)

        self.injections = injections
        self.injections.Nobs = len(list(posterior_samples_dictionary.keys())) # it's the number of GW events entering the analysis, used for the check Neff >= 4Nobs inside the injection class
        self.mass_priors = mass_priors
        #TODO: add check that the snr threshold is valid for this set of injections
        self.injections.update_cut(snr_cut=network_snr_threshold)
        self.cosmo = cosmo

        self.zprior_times_pxOmega_dict = {}
        self.pixel_indices_dictionary = {}
        self.samples_dictionary = {}
        self.samples_indices_dictionary = {}
        
        self.keys = []
                    
        for key, value in posterior_samples_dictionary.items():
            samples = load_posterior_samples(posterior_samples_dictionary[key],field=posterior_samples_field[key])
            skymap = gwcosmo.likelihood.skymap.skymap(skymap_dictionary[key])
            low_res_skyprob = hp.pixelfunc.ud_grade(skymap.prob, nside, order_in='NESTED', order_out='NESTED')
            low_res_skyprob = low_res_skyprob/np.sum(low_res_skyprob)
            
            pixelated_samples = make_pixel_px_function(samples, skymap, npixels=min_pixels, thresh=sky_area)
            nside_low_res = pixelated_samples.nside
            if nside_low_res > nside:
                raise ValueError(f'Low resolution nside {nside_low_res} is higher than high resolution nside {nside}. Try decreasing min_pixels for event {key}.')

            # identify which samples will be used to compute p(x|z,H0) for each pixel
            pixel_indices = pixelated_samples.indices
            samp_ind ={}
            for i,pixel_index in enumerate(pixel_indices):
                samp_ind[pixel_index] = pixelated_samples.identify_samples(pixel_index, minsamps=100)
                
            no_sub_pix_per_pixel = int(4**(np.log2(nside/nside_low_res)))

            # Get the coordinates of the hi-res pixel centres
            pixra, pixdec = ra_dec_from_ipix(nside, np.arange(hp.pixelfunc.nside2npix(nside)), nest=True)
            # compute the low-res index of each of them
            ipix = ipix_from_ra_dec(nside_low_res, pixra, pixdec, nest=True)

            print('Loading the redshift prior')
            zprior_times_pxOmega = np.zeros((len(pixel_indices),len(self.z_array)))
            for i,pixel_index in enumerate(pixel_indices):
                # Find the hi res indices corresponding to the current coarse pixel
                hi_res_pixel_indices = np.arange(hp.pixelfunc.nside2npix(nside))[np.where(ipix==pixel_index)[0]]
                # load pixels, weight by GW sky area, and combine
                for j, hi_res_index in enumerate(hi_res_pixel_indices):
                	zprior_times_pxOmega[i,:] +=  get_zprior(LOS_catalog, hi_res_index)*low_res_skyprob[hi_res_index]
            print(f"Identified {len(pixel_indices)*no_sub_pix_per_pixel} pixels in the galaxy catalogue which correspond to {key}'s {sky_area*100}% sky area")
            self.zprior_times_pxOmega_dict[key] = zprior_times_pxOmega
            self.pixel_indices_dictionary[key] = pixel_indices
            self.samples_dictionary[key] = samples
            self.samples_indices_dictionary[key] = samp_ind
            self.keys.append(key)
            
        LOS_catalog.close()
        print(self.keys)


    def log_likelihood_numerator_single_event(self,event_name):

        pixel_indices = self.pixel_indices_dictionary[event_name]
        samples = self.samples_dictionary[event_name]
        samp_ind = self.samples_indices_dictionary[event_name]
        zprior = self.zprior_times_pxOmega_dict[event_name]
                
        # set up KDEs for this value of the parameter to be analysed
        px_zOmegaparam = np.zeros((len(pixel_indices),len(self.z_array)))
        for i,pixel_index in enumerate(pixel_indices):

            z_samps,m1_samps,m2_samps = self.reweight_samps.compute_source_frame_samples(samples.distance[samp_ind[pixel_index]], samples.mass_1[samp_ind[pixel_index]], samples.mass_2[samp_ind[pixel_index]])

            zmin_temp = np.min(z_samps)*0.5
            zmax_temp = np.max(z_samps)*2.
            z_array_temp = np.linspace(zmin_temp,zmax_temp,100)

            kde,norm = self.reweight_samps.marginalized_redshift_reweight(z_samps,m1_samps,m2_samps)

            if norm != 0: # px_zOmegaH0 is initialized to 0
                px_zOmegaparam_interp = interp1d(z_array_temp,kde(z_array_temp),kind='cubic',bounds_error=False,fill_value=0)
                px_zOmegaparam[i,:] = px_zOmegaparam_interp(self.z_array)*norm
        
        # make p(s|z) have the same shape as p(x|z,Omega,param) and p(z|Omega,s)
        ps_z_array = np.tile(self.zrates(self.z_array),(len(pixel_indices),1))
        
        Inum_vals = np.sum(px_zOmegaparam*zprior*ps_z_array,axis=0)
        num = simpson(Inum_vals,self.z_array)

        return np.log(num)
        
    def log_likelihood_denominator_single_event(self):

        z_prior = interp1d(self.z_array,self.zprior_full_sky*self.zrates(self.z_array),bounds_error=False,fill_value=(0,(self.zprior_full_sky*self.zrates(self.z_array))[-1]))
        dz=np.diff(self.z_array)
        z_prior_norm = np.sum((z_prior(self.z_array)[:-1]+z_prior(self.z_array)[1:])*(dz)/2)
        injections = copy.deepcopy(self.injections) # Nobs is set in self.injection in the init

        # Update the sensitivity estimation with the new model
        injections.update_VT(self.cosmo,self.mass_priors,z_prior,z_prior_norm)
        Neff, Neff_is_ok, var = injections.calculate_Neff()
        if Neff_is_ok: # Neff >= 4*Nobs    
            log_den = np.log(injections.gw_only_selection_effect())
        else:
            print("Not enough Neff ({}) compared to Nobs ({}) for current mass-model {}, z-model {}, zprior_norm {}"
                  .format(Neff,injections.Nobs,self.mass_priors,z_prior,z_prior_norm))
            print("mass prior dict: {}, cosmo_prior_dict: {}".format(self.mass_priors_param_dict,self.cosmo_param_dict))
            print("returning infinite denominator")
            print("exit!")
            log_den = np.inf
            #sys.exit()
            
        return log_den, np.log(z_prior_norm)
                       
    def log_combined_event_likelihood(self):
        
        # carry norm to apply to numerator as well
        den_single, zprior_norm_log = self.log_likelihood_denominator_single_event()
        den = den_single*len(self.keys)
        
        num = 1.
        for event_name in self.keys:
            num += self.log_likelihood_numerator_single_event(event_name)-zprior_norm_log

        return num-den
        
    def log_likelihood(self):

        self.zrates.gamma = self.parameters['gamma']
        self.zrates.k = self.parameters['Madau_k']
        self.zrates.zp = self.parameters['Madau_zp']
        
        self.mass_priors_param_dict = {'alpha':self.parameters['alpha'], 'delta_m':self.parameters['delta_m'], 
                                         'mu_g':self.parameters['mu_g'], 'sigma_g':self.parameters['sigma_g'], 
                                         'lambda_peak':self.parameters['lambda_peak'],
                                         'alpha_1':self.parameters['alpha_1'], 
                                         'alpha_2':self.parameters['alpha_2'], 'b':self.parameters['b'], 
                                         'mminbh':self.parameters['mminbh'], 'mmaxbh':self.parameters['mmaxbh'], 
                                         'beta':self.parameters['beta'], 'alphans':self.parameters['alphans'],
                                         'mminns':self.parameters['mminns'], 'mmaxns':self.parameters['mmaxns']}

        self.mass_priors.update_parameters(self.mass_priors_param_dict)

        self.cosmo_param_dict = {'H0': self.parameters['H0'], 'Xi0': self.parameters['Xi0'], 'n': self.parameters['n']}
        self.cosmo.update_parameters(self.cosmo_param_dict)

        self.reweight_samps = reweight_posterior_samples(self.cosmo,self.mass_priors)
        
        return self.log_combined_event_likelihood()
        
    def __call__(self):
        return np.exp(self.log_likelihood())





#!/usr/bin/env python3

"""
Compute the coarse mth map for generating the LOS redshift prior

Freija Beirnaert
"""

import numpy as np
import healpy as hp
import sys

import gwcosmo

import logging
handler_out = logging.StreamHandler(stream=sys.stdout)
handler_err = logging.StreamHandler(stream=sys.stderr)
handler_err.setLevel(logging.ERROR)
logging.basicConfig(handlers=[handler_out, handler_err], level = logging.INFO)
logger = logging.getLogger(__name__)

def create_mth_map(outfile, catalog, band, nside, min_gal, pixel_index):
    """
    Generates a healpix map containing a magnitude threshold 
    based on the galaxies of a particular pixel in the sky.
    The result is saved as a fits file.

    Parameters
    ----------
    outfile : str
        path to the output file
    catalog : str
        galaxy catalogue
    band : str
        bbservation band of galaxy catalog (B,K,W1,bJ,u,g,r,i,z)
    nside : int
        healpy skymap nside
    min_gal : int
        the minimum number of galaxies in a pixel 
        for it not to be considered empty
    pixel_index: int
        healpy skymap pixel index for which mth is calculated
        if None, a full skymap is generated

    Returns
    -------
    None
    """
    
    logger.info(f"create {outfile}")
    full_catalog = gwcosmo.prior.catalog.load_catalog(catalog, band)

    npix = hp.pixelfunc.nside2npix(int(nside))
    m = np.zeros(npix)

    ipix = [pixel_index] # one pixel
    if pixel_index == None:
        ipix = np.arange(0, npix, 1) # full sky

    for pix in ipix:
        subcatalog = full_catalog.select_pixel(nside, pix, nested=True)
        mth = subcatalog.magnitude_thresh(band,min_gals=int(min_gal))
        m[pix] = mth

    hp.fitsfunc.write_map(outfile, m, nest=True, overwrite=True)




'''
Multi-parameter likelihood Module for GW events with 
electromagnetic counterpart

Tathagata Ghosh
'''

import numpy as np
from scipy.integrate import simpson, quad
from scipy.interpolate import interp1d
from gwcosmo.utilities.cosmology import standard_cosmology
from gwcosmo.likelihood.posterior_samples import *
from gwcosmo.likelihood.skymap import *
import gwcosmo
import bilby
import pickle
import copy
from scipy.stats import truncnorm
import sys

   
class MultipleEventLikelihoodEM(bilby.Likelihood):

    def __init__(self, counterpart_dictionary, injections, zrates, cosmo, mass_priors, posterior_samples_dictionary=None, posterior_samples_field=None, skymap_dictionary=None,  network_snr_threshold=12., post_los_dictionary=None, nsamps=1000, skymap_prior_distance="dlSquare", skymap_H0=70, skymap_Omega_m=0.3065):

        """
        Class to calculate log-likelihood on cosmological and population hyper-parameters.

        parameters
        ------------------

        counterpart_dictionary : dictionary
            Dictionary of elctomagnetic counterpart information of GW events.
                Structure of the dictionary: {"GW170817": {"redshift": [3017,166]/c,"ra_dec": [3.44602385, -0.40813555]}}
                                             [3017,166]/c=[mu/c, sigma/c] : mu, sigma from Gaussian distribution; c: speed of light in km/s.
        injections : injection object
            The injection object from gwcosmo to calculate selection effects.
                Pre-computed by using gwcosmo.gwcosmo.create_injections.py 
        zrates : gwcosmo.utilities.host_galaxy_merger_relations object
            Object of merger rate evolution with redshift.
        cosmo : gwcosmo.utilities.cosmology object
            Object of cosmological model.
        mass_priors : gwcosmo.prior.priors object
            Object of mass model: For example, BNS, NSBH_powerlaw.
        posterior_samples_dictionary : dictionary
            Dictionary to store the name of files containing posterior samples for GW events.
                Structure of the dictionary: {"GW170817": "GW170817_GWTC-1.hdf5"}
        skymap_dictionary : dictionary
            Dictionary to store the name of skymap files corresponding to GW events.
                Structure of the dictionary: {"GW170817": "GW170817_skymap.fits.gz"}
        network_snr_threshold : float
            Network SNR threshold of GW events which are used for analysis.
        post_los_dictionary : dictionary
            Dictionary to store the information whether posterior samples of GW events are conditioned over line of sight or not.
                Structure of the dictionary: {"GW170817": "True"}
        nsamps : dictionary
            Number of samples are considered from posterior samples when post_los is False.
        skymap_prior_distance : str 
            Prior used during construction of skymap. Default is "dlSquare". Other options are "UniformComoving" and "UniformComoving".
        skymap_H0 : float
            Assumed H0 while constructing skymap if skymap_prior_distance is "UniformComoving".
        skymap_Omega_m : float
            Assumed Omega_m (if Flat Lambda CDM cosmology)  while constructing skymap if skymap_prior_distance is "UniformComoving".
        """

        super().__init__(parameters={'H0': None, 'Xi0': None, 'n': None, 'gamma':None, 'Madau_k':None, 'Madau_zp':None, 'alpha':None, 'delta_m':None, 'mu_g':None, 'sigma_g':None, 'lambda_peak':None, 'alpha_1':None, 'alpha_2':None, 'b':None, 'mminbh':None, 'mmaxbh':None, 'beta':None, 'alphans':None, 'mminns':None, 'mmaxns':None})

        # em couterpart information
        self.counterpart_dictionary = counterpart_dictionary
        #redshift evolution model
        self.zrates = zrates

        #selection effect
        self.injections = injections
        self.injections.update_cut(snr_cut=network_snr_threshold)
        try:
            self.injections.Nobs = len(list(posterior_samples_dictionary.keys())) # it's the number of GW events entering the analysis, used for the check Neff >= 4Nobs inside the injection class
        except:
            self.injections.Nobs = len(list(skymap_dictionary.keys()))
        
        #mass distribution
        self.mass_priors = mass_priors

        #cosmology
        self.cosmo = cosmo

        # prior redshift distribution: uniform in comoving volume
        self.zprior = self.cosmo.p_z


        self.post_los_dictionary = post_los_dictionary
        self.sample_index_dictionary = {}
        self.nsamps = nsamps

        # counterpart information
        self.counterpart_zmin_zmax_dictionary = {}
        self.counterpart_pdf_dictionary = {}

        self.samples_dictionary = {}


        self.keys = []

        # likelihood in distance from skymap for em counterpart
        self.posterior_dl_skymap = {}
        self.skymap_prior_distance = skymap_prior_distance
        self.dlarray = {}


        # sample and skymap dictionary
        self.posterior_samples_dictionary = posterior_samples_dictionary
        self.skymap_dictionary = skymap_dictionary

        for key in self.counterpart_dictionary.keys():

            counterpart_muz, counterpart_sigmaz = self.counterpart_dictionary[key]["redshift"]
            zmin = counterpart_muz - 5*counterpart_sigmaz
            if zmin<0: zmin=0
            zmax = counterpart_muz + 5*counterpart_sigmaz
            a = (zmin - counterpart_muz) / counterpart_sigmaz
            b = (zmax - counterpart_muz) / counterpart_sigmaz
            counterpart_pdf = truncnorm(a,b,counterpart_muz,counterpart_sigmaz)
            self.counterpart_pdf_dictionary[key] = counterpart_pdf
            self.counterpart_zmin_zmax_dictionary[key] = np.array([zmin,zmax])
 
        if posterior_samples_dictionary is not None:
            for key, value in posterior_samples_dictionary.items():

                self.keys.append(key)

                samples = load_posterior_samples(posterior_samples_dictionary[key],field=posterior_samples_field[key])
                self.samples_dictionary[key] = samples
                if self.post_los_dictionary[key] is False:
                    ra_los, dec_los = self.counterpart_dictionary[key]["ra_dec"]
                    if type(self.nsamps) is dict and key in self.nsamps :
                        nsamp_event = int(self.nsamps [key])
                    else :
                        nsamp_event = int(self.nsamps)
                    sample_index, ang_rad_max = identify_samples_from_posterior (ra_los,dec_los,samples.ra,samples.dec,nsamp_event)
                    self.sample_index_dictionary [key] = sample_index
                    print (f"Considering {nsamp_event} samples around line of sight for {key}")

        elif skymap_dictionary is not None:

            for key, value in skymap_dictionary.items():
					
                skymap = gwcosmo.likelihood.skymap.skymap(skymap_dictionary[key])
                ra_los, dec_los = self.counterpart_dictionary[key]["ra_dec"]
                dlmin, dlmax, dlpost = skymap.lineofsight_posterior_dl(ra_los,dec_los)
                self.posterior_dl_skymap [key] = dlpost 
                if dlmin>0:
                    dl_array = np.linspace (dlmin,dlmax,10000)
                else :
                    dl_array = np.linspace (0,dlmax,10000)
                self.dlarray [key] = dl_array
                self.keys.append(key)

            if self.skymap_prior_distance=="UniformComoving" :
                zmin = 0 
                zmax = 10 
                cosmo_skymap = standard_cosmology (skymap_H0, skymap_Omega_m)
                z_array = np.linspace(zmin, zmax, 10000)
                dl_array = cosmo_skymap.dgw_z (z_array)
                z_prior_skymap = cosmo_skymap.p_z(z_array)
                self.dl_prior_skymap = interp1d (dl_array, z_prior_skymap)
        

    def log_likelihood_numerator_single_event_from_samples(self,event_name):
	
        samples = self.samples_dictionary[event_name]
        z_samps,m1_samps,m2_samps = self.reweight_samps.compute_source_frame_samples(samples.distance, samples.mass_1, samples.mass_2)
        if self.post_los_dictionary[event_name]:
            kde,norm = self.reweight_samps.marginalized_redshift_reweight(z_samps,m1_samps,m2_samps)		
        else :
            sample_index = self.sample_index_dictionary [event_name]
            kde,norm = self.reweight_samps.marginalized_redshift_reweight(z_samps[sample_index],m1_samps[sample_index],m2_samps[sample_index])
            
        redshift_bins = 1000
        zmin = self.cosmo.z_dgw(np.amin(samples.distance))*0.5
        zmax = self.cosmo.z_dgw(np.amax(samples.distance))*2.
        z_array_temp = np.linspace(zmin,zmax,redshift_bins)
        px_zOmegaH0_interp = interp1d(z_array_temp,kde(z_array_temp),kind='linear',bounds_error=False,fill_value=0)  # interpolation may produce some -ve values when kind='cubic'  
        num_x = lambda x : px_zOmegaH0_interp (x)*self.zrates(x)*self.counterpart_pdf_dictionary [event_name].pdf (x)
        zmin, zmax = self.counterpart_zmin_zmax_dictionary [event_name]
        num, _ = quad(num_x,zmin,zmax)

        return np.log(num*norm)
		
    def log_likelihood_numerator_single_event_from_skymap(self,event_name):

        redshift_bins = 10000
        zmin = self.cosmo.z_dgw(self.dlarray [event_name] [0])*0.5
        zmax = self.cosmo.z_dgw(self.dlarray [event_name] [-1])*2.
        z_array_temp = np.linspace(zmin,zmax,redshift_bins)

        dlarr_given_H0 = self.cosmo.dgw_z(z_array_temp)

        if self.skymap_prior_distance=="dlSquare" :
            likelihood_x_z_H0= self.posterior_dl_skymap[event_name].pdf(dlarr_given_H0)/dlarr_given_H0**2
        elif self.skymap_prior_distance=="Uniform" :
            likelihood_x_z_H0= self.posterior_dl_skymap[event_name].pdf(dlarr_given_H0)
        elif self.skymap_prior_distance=="UniformComoving" :
            likelihood_x_z_H0= self.posterior_dl_skymap[event_name].pdf(dlarr_given_H0)/self.dl_prior_skymap(dlarr_given_H0)
        likelihood_x_z_H0 /= simpson(likelihood_x_z_H0, z_array_temp)

        px_zOmegaH0_interp = interp1d(z_array_temp,likelihood_x_z_H0,kind='linear',bounds_error=False,fill_value=0)

        num_x = lambda x : px_zOmegaH0_interp(x)*self.zrates(x)*self.counterpart_pdf_dictionary[event_name].pdf(x) 
        zmin, zmax = self.counterpart_zmin_zmax_dictionary[event_name]
        num, _ = quad(num_x, zmin, zmax)

        return np.log(num)
	

    def log_likelihood_denominator_single_event(self):

        zmin = 0 
        zmax = 10 
        z_array = np.linspace(zmin, zmax, 10000)
        z_prior = interp1d(z_array,self.zprior(z_array)*self.zrates(z_array))
        dz=np.diff(z_array)
        z_prior_norm = np.sum((z_prior(z_array)[:-1]+z_prior(z_array)[1:])*(dz)/2)
        injections = copy.deepcopy(self.injections)  # Nobs is set in self.injection in the init
        
        # Update the sensitivity estimation with the new model
        injections.update_VT(self.cosmo,self.mass_priors,z_prior,z_prior_norm)
        Neff, Neff_is_ok, var = injections.calculate_Neff()
        if Neff_is_ok: # Neff >= 4*Nobs    
            log_den = np.log(injections.gw_only_selection_effect())
        else:
            print("Not enough Neff ({}) compared to Nobs ({}) for current mass-model {} and z-model {}".format(Neff,injections.Nobs,self.mass_priors,z_prior))
            print("mass prior dict: {}, cosmo_prior_dict: {}".format(self.mass_priors_param_dict,self.cosmo_param_dict))
            print("returning infinite denominator")
            print("exit!")
            log_den = np.inf
            #sys.exit()
            
        return log_den, np.log(z_prior_norm)

                       
    def log_combined_event_likelihood(self):
        
        num = 1.

        den_single, zprior_norm_log = self.log_likelihood_denominator_single_event()
        den = den_single*len(self.keys)

        if self.posterior_samples_dictionary is not None:
            for event_name in self.keys:
                num += self.log_likelihood_numerator_single_event_from_samples(event_name)-zprior_norm_log
        elif self.posterior_samples_dictionary is None and self.skymap_dictionary is not None :
            for event_name in self.keys:
                num += self.log_likelihood_numerator_single_event_from_skymap(event_name)-zprior_norm_log

        return num-den

    def log_likelihood(self):

        self.cosmo_param_dict = {'H0': self.parameters['H0'], 'Xi0': self.parameters['Xi0'], 'n': self.parameters['n']}
        self.cosmo.update_parameters(self.cosmo_param_dict)

        self.zrates.gamma = self.parameters['gamma']
        self.zrates.k = self.parameters['Madau_k']
        self.zrates.zp = self.parameters['Madau_zp']

        self.mass_priors_param_dict = {'alpha':self.parameters['alpha'], 'delta_m':self.parameters['delta_m'], 
                                         'mu_g':self.parameters['mu_g'], 'sigma_g':self.parameters['sigma_g'], 
                                         'lambda_peak':self.parameters['lambda_peak'],
                                         'alpha_1':self.parameters['alpha_1'], 
                                         'alpha_2':self.parameters['alpha_2'], 'b':self.parameters['b'], 
                                         'mminbh':self.parameters['mminbh'], 'mmaxbh':self.parameters['mmaxbh'], 
                                         'beta':self.parameters['beta'], 'alphans':self.parameters['alphans'],
                                         'mminns':self.parameters['mminns'], 'mmaxns':self.parameters['mmaxns']}

        self.mass_priors.update_parameters(self.mass_priors_param_dict)


        if self.posterior_samples_dictionary is not None:
            self.reweight_samps = reweight_posterior_samples(self.cosmo,self.mass_priors)
            return self.log_combined_event_likelihood()
            
        elif self.posterior_samples_dictionary is None and self.skymap_dictionary is not None :
            return self.log_combined_event_likelihood()	

    def __call__(self):
        return np.exp(self.log_likelihood())



import os
import glob
import pickle
import math
import array
import numpy as np

import h5py
import pkg_resources
import healpy as hp
import warnings

from ..likelihood.skymap import ipix_from_ra_dec
from ..utilities.cache import get_cachedir
from ..utilities.calc_kcor import calc_kcor

DEG2RAD = math.pi/180.0

color_names = {'B': None, 'K': None, 'u': 'u - r', 'g': 'g - r', 'r': 'g - r',
               'i': 'g - i', 'z': 'r - z', 'W1': None,'bJ': None}

# Taken from limits of K corrections from fig 4 of
# https://ui.adsabs.harvard.edu/abs/2010MNRAS.405.1409C/abstract
color_limits = {'u - r': [-0.1, 2.9], 'g - r': [-0.1, 1.9], 'g - i': [0, 3],
                'r - z': [0, 1.5], None : [-np.inf, np.inf]}

# istarmap.py for Python 3.8+
import multiprocessing.pool as mpp

def get_catalogfile(filename):
    if 'GWCOSMO_CATALOG_PATH' in os.environ.keys():
        file = os.path.join(os.environ['GWCOSMO_CATALOG_PATH'],filename)
    else:
        file = filename
    if os.path.exists(file):
        return file
    else:
        raise FileNotFoundError(f"Unable to locate {filename}. Make sure your $GWCOSMO_CATALOG_PATH is set")

def istarmap(self, func, iterable, chunksize=1):
    """starmap-version of imap
    """
    try:
        self._check_running()
    except AttributeError:
        if self._state != mpp.RUN:
            raise ValueError("Pool not running")
    if chunksize < 1:
        raise ValueError(
            "Chunksize must be 1+, not {0:n}".format(
                chunksize))

    task_batches = mpp.Pool._get_tasks(func, iterable, chunksize)
    try:
        result = mpp.IMapIterator(self)
    except TypeError:
        result = mpp.IMapIterator(self._cache)

    self._taskqueue.put(
        (
            self._guarded_task_generation(result._job,
                                          mpp.starmapstar,
                                          task_batches),
            result._set_length
        ))
    return (item for chunk in result for item in chunk)


mpp.Pool.istarmap = istarmap

def load_catalog(name, band):
    # Load raw catalog
    if name == 'GLADE':
        cat = OldStyleGLADE(band=band)
    elif name == 'DES':
        cat = OldStyleDES(band=band)
    elif name == 'DESI':
        cat = OldStyleDESI(band=band)
    elif name == 'GLADE+':
        cat = OldStyleGLADEPlus(band=band)
    else:
        raise ValueError(f"Unable catalog {name}")
    if band is not None:
        cat = cat.remove_missing_magnitudes(band)
    return cat

def load_catalog_from_opts(opts):
    name = opts.catalog
    band = opts.catalog_band
    return load_catalog(name, band)

from abc import ABC, abstractmethod

class GalaxyCatalog:
    """
    Interface for a generic galaxy catalog
    """
    colnames = {'z','sigmaz','ra','dec'}

    def __init__(self, data = None, name = 'Unknown Catalog',
                 supported_bands=None, cachedir=None):
        self.data = data
        self.name = name
        self.supported_bands = supported_bands
        if supported_bands is not None:
            self.colnames.union([f'm_{band}' for band in supported_bands])
        # Cache for pixel index to array index lookup
        self.pixmap = {}

    def __getitem__(self, *args, **kwargs):
        return self.data.__getitem__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        return self.data.__setitem__(*args, **kwargs)

    def __len__(self):
        if self.data is None:
            return 0
        else:
            return len(self.data)

    def build_pixel_index_file(self, nside, cachedir=None, nested=True):
        if cachedir is None:
            cachedir = get_cachedir()

        filepath = self._cachefile(nside, nested=nested, cachedir=cachedir)
        pixlists = pixelate(self, nside, nested=nested)
        with open(filepath,'wb') as idxfile:
            pickle.dump(pixlists, idxfile)

    def _cachefile(self, nside, nested=True, cachedir=None):
        if cachedir is None:
            cachedir = get_cachedir()

        return os.path.join(cachedir,
                           '_'.join([self.name, str(nside), str(nested), 'pixidx.pkl'] )
                           )

    def read_pixel_index_cache(self, nside, cachedir=None):
        cachefile = self._cachefile(nside, cachedir=cachedir)
        if os.path.exists(cachefile):
            try:
                self.pixmap[nside] = pickle.load(open(cachefile,'rb'))
                return True
            except:
                print(f"Warning, unable to open pixel cache file {cachefile}, possible corrupted file")
                return False
        else:
            return False

    def clean_cache(self, mtime, cachedir=None):
        """
        Remove any index cache files older than mtime
        """
        if cachedir is None:
            cachedir = get_cachedir()
        for f in glob.glob(cachedir+'/'+self.name+'_*'):
            try:
                if os.path.getmtime(f) < mtime:
                    os.remove(f)
            except FileNotFoundError:
                # Here in case a parallel job has removed the file
                pass

    def magnitude_thresh(self, band, ra=None, dec=None, min_gals=10, method="median"):
        """
        Return the magnitude threshold for a specific
        sky position if specified. The magnitude threshold is
        estimated as the median of the observed magnitudes
        If fewer than min_gals galaxies exist then return inf
        TODO: Currently this doesn't select based on ra,dec
        """
        assert (ra is None and dec is None) or (ra is not None and dec is not None)

        if len(self) < min_gals:
            mth = np.inf
        else:
            if method == "median":
                m = self.get_magnitudes(band)
                mth = np.median(m[np.isfinite(m)])
            #TODO other methods
            else:
                warnings.warn(f"Method {method} not recognized, using median method")
                m = self.get_magnitudes(band)
                mth = np.median(m[np.isfinite(m)])
        return mth

    def select_pixel(self, nside, pixel_index, nested=True):
        """
        Keep only galaxies in the desired healpix pixel indices
        """
        # Try to load an index file, and create one if not
        if not self.read_pixel_index_cache(nside):
            # Try to make the index file
            self.build_pixel_index_file(nside, nested=nested)
            # Check if it can be read
            if not self.read_pixel_index_cache(nside):
                # If not, build it here
                print('No cache file found, generating pixel index (may take some time)')
                self.pixmap[nside] = pixelate(self, nside, nested=nested)

        pixmap = self.pixmap[nside]

        idx = pixmap[pixel_index]
        new = GalaxyCatalog(data = self.data[idx], name = self.name+f'_nside{nside}_pixel{pixel_index}',
                            supported_bands = self.supported_bands)
        return new

    def idx2pixdict(self, nside, idx, nested=True):
        ra, dec = self['ra'][idx], self['dec'][idx]
        ipix = ipix_from_ra_dec(nside, ra, dec, nest=nested)
        return (ipix, idx)

    def get_magnitudes(self, band):
        """
        Get magnitudes for a particular band
        """
        assert band in self.supported_bands, f"error: catalog doesn't support {band}-band magnitudes"
        return self[f'm_{band}']

    def get_color(self, band):
        """
        Return color index for K corrections
        """
        Kcorr_bands = {'B': 'B', 'K': 'K','bJ':'bJ','u': 'r', 'g': 'r', 'r': 'g', 'i': 'g', 'z': 'r'}
        Kcorr_signs = {'B': 1, 'K': 1,'bJ':1 ,'u': 1, 'g': 1, 'r': -1, 'i': -1, 'z': -1}
        if not band in Kcorr_bands.keys():
            print('Band {band} not supported for color-based K corrections')
            color = np.zeros(len(self))
        else:
            m = self.get_magnitudes(band)
            m_K = self.get_magnitudes(Kcorr_bands[band])
            color = Kcorr_signs[band]*(m - m_K)
        return color

    def apply_redshift_cut(self, zcut):
        idx = np.where(((self['z']-3*self['sigmaz']) <= zcut))
        return GalaxyCatalog(data = self.data[idx], name = self.name+f'_zcut{zcut}',
                             supported_bands = self.supported_bands)

    def remove_missing_magnitudes(self, band):
        m = self.get_magnitudes(band)
        idx = np.where(np.isfinite(m))
        return GalaxyCatalog(data = self.data[idx], name = self.name+f'_valid_{band}_mags',
                             supported_bands = self.supported_bands)

    def apply_color_limit(self, band, cmin, cmax):
        """
        Apply the given color limits to the given band.
        Will filter out any galaxies whose magnitude is missing in given band.
        """
        if band == 'W1':
            print('Not applying color limits for W1 band, as we use the z-dependent k-correction')
            return self
        idx = np.where((cmin <= self.get_color(band)) & (cmax >= self.get_color(band)))
        return GalaxyCatalog(data = self.data[idx], name = self.name+f'_climit_{band}_{cmin}_{cmax}',
                             supported_bands = self.supported_bands)

    def apply_magnitude_limit(self, band, magmax):
        idx = np.where(self.get_magnitudes(band) <= magmax)
        return GalaxyCatalog(data = self.data[idx], name = self.name+f'_mlimit_{band}_{magmax}',
                             supported_bands = self.supported_bands)

    def get_k_correction(self, band, z, color_name, color_value):
        "Apply K-correction"
        #https://arxiv.org/pdf/1709.08316.pdf
        if band == 'W1':
            k_corr = -1*(4.44e-2+2.67*z+1.33*(z**2.)-1.59*(z**3.)) #From Maciej email
            return k_corr
        elif band == 'K':
             # https://iopscience.iop.org/article/10.1086/322488/pdf 4th page lhs
            to_ret=-6.0*np.log10(1+z)
            to_ret[z>0.3]=-6.0*np.log10(1+0.3)
            return -6.0*np.log10(1+z)
        elif band == 'bJ':
            # Fig 5 caption from https://arxiv.org/pdf/astro-ph/0111011.pdf
            # Note that these corrections also includes evolution corrections
            return (z+6*np.power(z,2.))/(1+15*np.power(z,3.))
        else:
            try:
                kcor = calc_kcor(band, z, color_name, color_value)
                return kcor
            except:
                print(f'Cannot calculate k-corrections for band {band} color name {color_name}')
                print('Will return 0 k corrections')
                return np.zeros(len(z))

def newarr():
    return array.array('Q')

def pixelate(cat, nside, allowed_pixels=None, nested=True):
    from multiprocessing import Pool
    from collections import defaultdict
    from tqdm import tqdm
    Nprocs = 2 # higher values use more RAM and disk bandwidth.

    # Arrays of unsigned 64-bit integers for the galaxy indices
    pixlists = defaultdict(newarr)

    with Pool(processes=Nprocs) as pool:

        for (pixidx, galidx) in \
            pool.istarmap(cat.idx2pixdict,
                          ( [nside, idx] for idx in range(len(cat)) ),
                          chunksize=1000000
                         ):
                pixlists[pixidx].append(galidx)
    for k,v in pixlists.items():
        pixlists[k]=np.sort(v)
    return pixlists


class OldStyleCatalog(GalaxyCatalog):
    """
    Catalog in the old GWCosmo format. Must have been
    preprocessed into HDF5 files.
    """
    filename = None
    def __init__(self,
                     catalog_file=None,
                     name = None):

        self.filename = get_catalogfile(catalog_file)

        super().__init__(supported_bands = self.supported_bands, name = name)
        self.populate()
        self.clean_cache(os.path.getmtime(self.filename))

    def populate(self):
        """
        This is a separate step to load the data
        """
        f = h5py.File(self.filename, 'r')
        names = []
        for n in self.colnames:
            if n in f:
                names.append(n)
            else:
                print(f'Unable to find column for {n}-band')
        self.data = np.rec.fromarrays([f[n] for n in names], names = names)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['data']
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self.populate()


class OldStyleDESI(OldStyleCatalog):
    supported_bands = {'W1'}
    def __init__(self, catalog_file = 'DESI.hdf5', band='W1'):
        print('WARNING: DESI catalog is not fully supported yet')
        self.colnames = set(self.colnames).union([f'm_{b}' for b in self.supported_bands])
        super().__init__(catalog_file = catalog_file, name = 'DESI')

class OldStyleGLADEPlus(OldStyleCatalog):
    supported_bands = {'B', 'K', 'W1','bJ'}
    def __init__(self, catalog_file = 'glade+.hdf5', band='bJ'):
        self.colnames = set(self.colnames).union([f'm_{b}' for b in self.supported_bands])
        super().__init__(catalog_file = catalog_file, name = 'GladePlus')

class OldStyleGLADE(OldStyleCatalog):
    supported_bands = {'B','K'}
    def __init__(self, catalog_file = 'glade.hdf5', band='B'):
        self.colnames = set(self.colnames).union([f'm_{b}' for b in self.supported_bands])
        super().__init__(catalog_file = catalog_file, name = 'Glade')

class OldStyleDES(OldStyleCatalog):
    supported_bands = {'g','i','r','z'}
    def __init__(self, catalog_file = 'des.hdf5', band='G'):
        self.colnames = set(self.colnames).union([f'm_{b.lower()}' for b in self.supported_bands])
        super().__init__(catalog_file = catalog_file, name = 'DES')




'''
This module collects analytical and numerical probability density functions.
'''

import numpy as _np
from scipy.stats import truncnorm as _truncnorm
import copy as _copy
from scipy.interpolate import interp1d as _interp1d
from scipy.special import erf as _erf
from scipy.special import logsumexp as _logsumexp
from scipy.integrate import cumtrapz as _cumtrapz

def _S_factor(mass, mmin,delta_m):
    '''
    This function return the value of the window function defined as Eqs B6 and B7 of https://arxiv.org/pdf/2010.14533.pdf

    Parameters
    ----------
    mass: np.array or float
        array of x or masses values
    mmin: float or np.array (in this case len(mmin) == len(mass))
        minimum value of window function
    delta_m: float or np.array (in this case len(delta_m) == len(mass))
        width of the window function

    Returns
    -------
    Values of the window function
    '''

    if not isinstance(mass,_np.ndarray):
        mass = _np.array([mass])

    to_ret = _np.ones_like(mass)
    if delta_m == 0:
        return to_ret

    mprime = mass-mmin

    # Defines the different regions of thw window function ad in Eq. B6 of  https://arxiv.org/pdf/2010.14533.pdf
    select_window = (mass>mmin) & (mass<(delta_m+mmin))
    select_one = mass>=(delta_m+mmin)
    select_zero = mass<=mmin

    effe_prime = _np.ones_like(mass)

    # Definethe f function as in Eq. B7 of https://arxiv.org/pdf/2010.14533.pdf
    effe_prime[select_window] = _np.exp(_np.nan_to_num((delta_m/mprime[select_window])+(delta_m/(mprime[select_window]-delta_m))))
    to_ret = 1./(effe_prime+1)
    to_ret[select_zero]=0.
    to_ret[select_one]=1.
    return to_ret

def get_PL_norm(alpha,minv,maxv):
    '''
    This function returns the powerlaw normalization factor

    Parameters
    ----------
    alpha: float
        Powerlaw slope
    min_pl: float
        lower cutoff
    max_pl: float
        upper cutoff
    '''

    # Get the PL norm as in Eq. 24 on the tex document
    if alpha == -1:
        return _np.log(maxv/minv)
    else:
        return (_np.power(maxv,alpha+1) - _np.power(minv,alpha+1))/(alpha+1)

def get_gaussian_norm(mu,sigma,min,max):
    '''
    This function returns the gaussian normalization factor

    Parameters
    ----------
    mu: float
        mean of the gaussian
    sigma: float
        standard deviation of the gaussian
    min_pl: float
        lower cutoff
    max_pl: float
        upper cutoff
    '''

    # Get the gaussian norm as in Eq. 28 on the tex document
    max_point = (max-mu)/(sigma*_np.sqrt(2.))
    min_point = (min-mu)/(sigma*_np.sqrt(2.))
    return 0.5*_erf(max_point)-0.5*_erf(min_point)


class SmoothedProb(object):
    '''
    Class for smoothing the low part of a PDF. The smoothing follows Eq. B7 of
    2010.14533.

    Parameters
    ----------
    origin_prob: class
        Original prior class to smooth from this module
    bottom: float
        minimum cut-off. Below this, the window is 0.
    bottom_smooth: float
        smooth factor. The smoothing acts between bottom and bottom+bottom_smooth
    '''

    def __init__(self,origin_prob,bottom,bottom_smooth):

        self.origin_prob = _copy.deepcopy(origin_prob)
        self.bottom_smooth = bottom_smooth
        self.bottom = bottom
        self.maximum=self.origin_prob.maximum
        self.minimum=self.origin_prob.minimum

        # Find the values of the integrals in the region of the window function before and after the smoothing
        int_array = _np.linspace(bottom,bottom+bottom_smooth,1000)
        integral_before = _np.trapz(self.origin_prob.prob(int_array),int_array)
        integral_now = _np.trapz(self.prob(int_array),int_array)

        self.integral_before = integral_before
        self.integral_now = integral_now
        # Renormalize the the smoother function.
        self.norm = 1 - integral_before + integral_now - self.origin_prob.cdf(bottom)

        x_eval = _np.logspace(_np.log10(bottom),_np.log10(bottom+bottom_smooth),1000)
        cdf_numeric = _cumtrapz(self.prob(x_eval),x_eval)
        self.cached_cdf_window = _interp1d(x_eval[:-1:],cdf_numeric,fill_value='extrapolate',bounds_error=False,kind='linear')

        
    def prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        return _np.exp(self.log_prob(x))

    def log_prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        # Return the window function
        window = _S_factor(x, self.bottom,self.bottom_smooth)

        if hasattr(self,'norm'):
            prob_ret =self.origin_prob.log_prob(x)+_np.log(window)-_np.log(self.norm)
        else:
            prob_ret =self.origin_prob.log_prob(x)+_np.log(window)

        return prob_ret

    def log_conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array
            Value at which compute the probability
        a: np.array
            New lower boundary
        b: np.array
            New upper boundary
        """

        to_ret = self.log_prob(x)
        # Find the new normalization in the new interval
        new_norm = self.cdf(b)-self.cdf(a)
        # Apply the new normalization and put to zero all the values above/below the interval
        wok = _np.where(new_norm > 0)[0]
        if len(wok)>0: # the pdf can be normalized
            to_ret[wok] -= _np.log(new_norm[wok])
        wnull = _np.where(new_norm <= 0)[0]
        if len(wnull)>0:            
            to_ret[wnull] = -_np.inf

        to_ret[(x<a) | (x>b)] = -_np.inf

        return to_ret

    def cdf(self,x):
        """
        Returns the cumulative density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the cumulative
        """

        to_ret = _np.ones_like(x)
        to_ret[x<self.bottom] = 0.
        to_ret[(x>=self.bottom) & (x<=(self.bottom+self.bottom_smooth))] = self.cached_cdf_window(x[(x>=self.bottom) & (x<=(self.bottom+self.bottom_smooth))])
        to_ret[x>=(self.bottom+self.bottom_smooth)]=(self.integral_now+self.origin_prob.cdf(
        x[x>=(self.bottom+self.bottom_smooth)])-self.origin_prob.cdf(self.bottom+self.bottom_smooth))/self.norm

        return to_ret

    def conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array
            Value at which compute the probability
        a: np.array
            New lower boundary
        b: np.array
            New upper boundary
        """

        return _np.exp(self.log_conditioned_prob(x,a,b))

class PowerLaw_math(object):
    """
    Class for a powerlaw probability :math:`p(x) \\propto x^{\\alpha}` defined in
    [a,b]

    Parameters
    ----------
    alpha: float
        Powerlaw slope
    min_pl: float
        lower cutoff
    max_pl: float
        upper cutoff
    """

    def __init__(self,alpha,min_pl,max_pl):

        self.minimum = min_pl
        self.maximum = max_pl
        self.min_pl = min_pl
        self.max_pl = max_pl
        self.alpha = alpha

        # Get the PL norm and as Eq. 24 on the paper
        self.norm = get_PL_norm(alpha,min_pl,max_pl)

    def prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        return _np.exp(self.log_prob(x))

    def log_prob(self,x):
        """
        Returns the logarithm of the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        to_ret = self.alpha*_np.log(x)-_np.log(self.norm)
        to_ret[(x<self.min_pl) | (x>self.max_pl)] = -_np.inf

        return to_ret

    def log_conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        a: np.array or float
            New lower boundary
        b: np.array or float
            New upper boundary
        """

        norms = get_PL_norm(self.alpha,a,b)
        to_ret = self.alpha*_np.log(x)-_np.log(norms)
        to_ret[(x<a) | (x>b)] = -_np.inf

        return to_ret

    def cdf(self,x):
        """
        Returns the cumulative density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the cumulative
        """

        # Define the cumulative density function, see  Eq. 24 to see the integral form

        if self.alpha == -1:
            to_ret = _np.log(x/self.min_pl)/self.norm
        else:
            to_ret =((_np.power(x,self.alpha+1)-_np.power(self.min_pl,self.alpha+1))/(self.alpha+1))/self.norm

        to_ret *= (x>=self.min_pl)

        if hasattr(x, "__len__"):
            to_ret[x>self.max_pl]=1.
        else:
            if x>self.max_pl : to_ret=1.

        return to_ret

    def conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        a: np.array or float
            New lower boundary
        b: np.array or float
            New upper boundary
        """

        return _np.exp(self.log_conditioned_prob(x,a,b))



class Truncated_Gaussian_math(object):
    """
    Class for a truncated gaussian in
    [a,b]

    Parameters
    ----------
    mu: float
        mean of the gaussian
    sigma: float
        standard deviation of the gaussian
    min_g: float
        lower cutoff
    max_g: float
        upper cutoff
    """

    def __init__(self,mu,sigma,min_g,max_g):

        self.minimum = min_g
        self.maximum = max_g
        self.max_g=max_g
        self.min_g=min_g
        self.mu = mu
        self.sigma=sigma

        # Find the gaussian normalization as in Eq. 28 in the tex document
        self.norm = get_gaussian_norm(mu,sigma,min_g,max_g)

    def prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        return _np.exp(self.log_prob(x))


    def log_prob(self,x):
        """
        Returns the logarithm of the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        to_ret = -_np.log(self.sigma)-0.5*_np.log(2*_np.pi)-0.5*_np.power((x-self.mu)/self.sigma,2.)-_np.log(self.norm)
        to_ret[(x<self.min_g) | (x>self.max_g)] = -_np.inf

        return to_ret

    def log_conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        a: np.array or float
            New lower boundary
        b: np.array or float
            New upper boundary
        """

        norms = get_gaussian_norm(self.mu,self.sigma,a,b)
        to_ret = -_np.log(self.sigma)-0.5*_np.log(2*_np.pi)-0.5*_np.power((x-self.mu)/self.sigma,2.)-_np.log(norms)
        to_ret[(x<a) | (x>b)] = -_np.inf

        return to_ret


    def cdf(self,x):
        """
        Returns the cumulative density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the cumulative
        """

        # Define the cumulative density function as in Eq. 28 on the paper to see the integral form

        max_point = (x-self.mu)/(self.sigma*_np.sqrt(2.))
        min_point = (self.min_g-self.mu)/(self.sigma*_np.sqrt(2.))

        to_ret = (0.5*_erf(max_point)-0.5*_erf(min_point))/self.norm

        to_ret *= (x>=self.min_g)

        if hasattr(x, "__len__"):
            to_ret[x>self.max_g]=1.
        else:
            if x>self.max_g : to_ret=1.

        return to_ret

    def conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        a: np.array or float
            New lower boundary
        b: np.array or float
            New upper boundary
        """

        return _np.exp(self.log_conditioned_prob(x,a,b))



class PowerLawGaussian_math(object):
    """
    Class for a powerlaw probability plus gausian peak
    :math:`p(x) \\propto (1-\\lambda)x^{\\alpha}+\\lambda \\mathcal{N}(\\mu,\\sigma)`. Each component is defined in
    a different interval

    Parameters
    ----------
    alpha: float
        Powerlaw slope
    min_pl: float
        lower cutoff
    max_pl: float
        upper cutoff
    lambda_g: float
        fraction of prob coming from gaussian peak
    mean_g: float
        mean for the gaussian
    sigma_g: float
        standard deviation for the gaussian
    min_g: float
        minimum for the gaussian component
    max_g: float
        maximim for the gaussian component
    """

    def __init__(self,alpha,min_pl,max_pl,lambda_g,mean_g,sigma_g,min_g,max_g):

        self.minimum = _np.min([min_pl,min_g])
        self.maximum = _np.max([max_pl,max_g])

        self.lambda_g=lambda_g

        self.pl= PowerLaw_math(alpha,min_pl,max_pl)
        self.gg = Truncated_Gaussian_math(mean_g,sigma_g,min_g,max_g)


    def prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        # Define the PDF as in Eq. 36-37-38 on on the tex document
        return _np.exp(self.log_prob(x))

    def cdf(self,x):
        """
        Returns the cumulative density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the cumulative
        """

        return (1-self.lambda_g)*self.pl.cdf(x)+self.lambda_g*self.gg.cdf(x)

    def conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        a: np.array or float
            New lower boundary
        b: np.array or float
            New upper boundary
        """

        return _np.exp(self.log_conditioned_prob(x,a,b))

    def log_prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        # Define the PDF as in Eq. 36-37-38 on on the tex document
        return _np.logaddexp(_np.log1p(-self.lambda_g)+self.pl.log_prob(x),_np.log(self.lambda_g)+self.gg.log_prob(x))

    def log_conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        a: np.array or float
            New lower boundary
        b: np.array or float
            New upper boundary
        """

        return  _np.logaddexp(_np.log1p(-self.lambda_g)+self.pl.log_conditioned_prob(x,a,b),_np.log(self.lambda_g)+self.gg.log_conditioned_prob(x,a,b))


class PowerLawDoubleGaussian_math(object):
    """
    Class for a powerlaw probability plus gausian peak
    :math:`p(x) \\propto (1-\\lambda)x^{\\alpha}+\\lambda \\lambda_1 \\mathcal{N}(\\mu_1,\\sigma_1)+\\lambda (1-\\lambda_1) \\mathcal{N}(\\mu_2,\\sigma_2)`.
    Each component is defined ina different interval

    Parameters
    ----------
    alpha: float
        Powerlaw slope
    min_pl: float
        lower cutoff
    max_pl: float
        upper cutoff
    lambda_g: float
        fraction of prob coming in any gaussian peak
    lambda_g_low: float
        fraction of prob in lower gaussian peak
    mean_g_low: float
        mean for the gaussian
    sigma_g_low: float
        standard deviation for the gaussian# Define the PDF as in Eq. 37 on on the tex document
    mean_g_high: float
        mean for the gaussian
    sigma_g_high: float
        standard deviation for the gaussian
    min_g: float
        minimum for the gaussian components
    max_g: float
        maximim for the gaussian components
    """

    def __init__(self,alpha,min_pl,max_pl,lambda_g, lambda_g_low,
    mean_g_low,sigma_g_low,mean_g_high,sigma_g_high,min_g,max_g):

        self.minimum = _np.min([min_pl,min_g])
        self.maximum = _np.max([max_pl,max_g])

        self.lambda_g = lambda_g
        self.lambda_g_low = lambda_g_low

        self.pl= PowerLaw_math(alpha,min_pl,max_pl)
        self.gg_low = Truncated_Gaussian_math(mean_g_low,sigma_g_low,min_g,max_g)
        self.gg_high = Truncated_Gaussian_math(mean_g_high,sigma_g_high,min_g,max_g)

    def prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        return _np.exp(self.log_prob(x))


    def log_prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        # Define the PDF as in Eq. 44-45-46 on the tex document

        pl_part = _np.log1p(-self.lambda_g)+self.pl.log_prob(x)
        g_low = self.gg_low.log_prob(x)+_np.log(self.lambda_g)+_np.log(self.lambda_g_low)
        g_high = self.gg_high.log_prob(x)+_np.log(self.lambda_g)+_np.log1p(-self.lambda_g_low)

        return _logsumexp(_np.stack([pl_part,g_low,g_high]),axis=0)

    def log_conditioned_prob(self,x,a,b):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        # Define the PDF as in Eq.  44-45-46  on the tex document

        pl_part = _np.log1p(-self.lambda_g)+self.pl.log_conditioned_prob(x,a,b)
        g_low = self.gg_low.log_conditioned_prob(x,a,b)+_np.log(self.lambda_g)+_np.log(self.lambda_g_low)
        g_high = self.gg_high.log_conditioned_prob(x,a,b)+_np.log(self.lambda_g)+_np.log1p(-self.lambda_g_low)

        return _logsumexp(_np.stack([pl_part,g_low,g_high]),axis=0)

    def cdf(self,x):
        """
        Returns the cumulative density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the cumulative
        """

        pl_part = (1-self.lambda_g)*self.pl.cdf(x)
        g_part =self.gg_low.cdf(x)*self.lambda_g*self.lambda_g_low+self.gg_high.cdf(x)*self.lambda_g*(1-self.lambda_g_low)

        return pl_part+g_part

    def conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        a: np.array or float
            New lower boundary
        b: np.array or float
            New upper boundary
        """

        return _np.exp(self.log_conditioned_prob(x,a,b))



class BrokenPowerLaw_math(object):
    """
    Class for a broken powerlaw probability
    :math:`p(x) \\propto x^{\\alpha}` if :math:`min<x<b(max-min)`, :math:`p(x) \\propto x^{\\beta}` if :math:`b(max-min)<x<max`.

    Parameters
    ----------
    alpha_1: float
        Powerlaw slope for first component
    alpha_2: float
        Powerlaw slope for second component
    min_pl: float
        lower cutoff
    max_pl: float
        upper cutoff
    b: float
        fraction in [0,1] at which the powerlaw breaks
    """

    def __init__(self,alpha_1,alpha_2,min_pl,max_pl,b):

        self.minimum = min_pl
        self.maximum = max_pl

        self.min_pl = min_pl
        self.max_pl = max_pl

        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2

        # Define the breaking point
        self.break_point = min_pl+b*(max_pl-min_pl)
        self.b=b

        # Initialize the single powerlaws
        self.pl1=PowerLaw_math(alpha_1,min_pl,self.break_point)
        self.pl2=PowerLaw_math(alpha_2,self.break_point,max_pl)

        # Define the broken powerlaw as in Eq. 39-40-41 on the tex document
        self.new_norm=(1+self.pl1.prob(_np.array([self.break_point]))/self.pl2.prob(_np.array([self.break_point])))

    def prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        # Define the PDF as in Eq. 39-40-41 on the tex document
        return _np.exp(self.log_prob(x))


    def log_prob(self,x):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        # Define the PDF as in Eq. 39-40-41 on the tex document

        to_ret = _np.logaddexp(self.pl1.log_prob(x),self.pl2.log_prob(x)+self.pl1.log_prob(_np.array([self.break_point]))
        -self.pl2.log_prob(_np.array([self.break_point])))-_np.log(self.new_norm)
        return to_ret

    def log_conditioned_prob(self,x,a,b):
        """
        Returns the probability density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        """

        # Define the PDF as in Eq. 39-40-41 on the tex document

        to_ret = _np.logaddexp(self.pl1.log_conditioned_prob(x,a,b),self.pl2.log_conditioned_prob(x,a,b)
        +self.pl1.log_prob(_np.array([self.break_point]))-self.pl2.log_prob(_np.array([self.break_point])))-_np.log(self.new_norm)

        return to_ret

    def cdf(self,x):
        """
        Returns the cumulative density function normalized

        Parameters
        ----------
        x: np.array or float
            Value at which compute the cumulative
        """

        return (self.pl1.cdf(x)+self.pl2.cdf(x)*(self.pl1.prob(_np.array([self.break_point]))/self.pl2.prob(_np.array([self.break_point]))))/self.new_norm

    def conditioned_prob(self,x,a,b):
        """
        Returns the conditional probability between two new boundaries [a,b]

        Parameters
        ----------
        x: np.array or float
            Value at which compute the probability
        a: np.array or float
            New lower boundary
        b: np.array or float
            New upper boundary
        """

        return _np.exp(self.log_conditioned_prob(x,a,b))



#!/usr/bin/env python3

"""
Compute the coarse galaxy_norm map for generating the LOS redshift prior

Freija Beirnaert
"""

import numpy as np
import healpy as hp
import sys

import gwcosmo
from scipy.integrate import quad, dblquad
from scipy.stats import norm, truncnorm
from scipy.interpolate import interp1d
from gwcosmo.utilities.luminosity_function import M_Mobs,M_mdl
from gwcosmo.prior.catalog import color_names, color_limits

import logging
handler_out = logging.StreamHandler(stream=sys.stdout)
handler_err = logging.StreamHandler(stream=sys.stderr)
handler_err.setLevel(logging.ERROR)
logging.basicConfig(handlers=[handler_out, handler_err], level = logging.INFO)
logger = logging.getLogger(__name__)

def calc_norm(catalog, band, nside, pixel_index, zmax, zcut, mth, cosmology, schech_Mmax, apply_Kcorr):
    """
    Calculate the average normalization factor based on the galaxies 
    of a particular healpy pixel in the sky.
    The resulting number is used in the LOS redshift prior calculation.

    Parameters
    ----------
    catalog : str
        galaxy catalogue
    band : str
        bbservation band of galaxy catalog (B,K,W1,bJ,u,g,r,i,z)
    nside : int
        healpy skymap nside
    pixel_index: int
        healpy skymap pixel index for which mth is calculated
        if None, a full skymap is generated
    zmax: float
        upper redshift limit for integrals
    zcut: float
        hard redshift cut to apply to the galaxy catalogue
    mth: float
        value of the magnitude threshold
    cosmology: gwcosmo.utilities.cosmology.standard_cosmology object
        cosmological model
    schech_Mmax : float
        value for Mmax of schechter function for a given band   
    apply_Kcorr: bool
        Apply K-corrections
    
    Returns
    -------
    float
    """

    subcatalog = catalog.select_pixel(nside, pixel_index, nested=True)
    color_limit = 0.
    if apply_Kcorr:
        color_limit = color_limits[color_names[band]]
    else:
        color_limit = [-np.inf,np.inf]
    subcatalog = subcatalog.apply_redshift_cut(zcut).apply_color_limit(band, color_limit[0], color_limit[1]).apply_magnitude_limit(band, mth)

    zs = subcatalog['z'] 
    ms = subcatalog.get_magnitudes(band)
    sigmazs = subcatalog['sigmaz']
    colors = subcatalog.get_color(band)
    Mmax = M_Mobs(cosmology.H0,schech_Mmax)
    
    galaxy_norm = 0.
    N_in3sigma = 0
    logger.info(f"nside {nside} pixel {pixel_index}: The initial number of galaxies in this pixel is {len(zs)}")
    for i in range(len(zs)):
        z_array = np.linspace(zs[i]-5*sigmazs[i],zs[i]+5*sigmazs[i],50) # redshift array for truncnorm interpolations in redshiftuncertainties
        # set redshift limits so that galaxies can't have negative z
        low_z_lim, high_z_lim = (0 - zs[i])/sigmazs[i], (zmax - zs[i])/sigmazs[i]
        # Find the effective number of galaxies below zcut. If zcut=zmax then galaxy_norm=Ngal.
        # Takes into account galaxies whose z distribution crosses zcut.
        if low_z_lim > -3:
            N_in3sigma += 1
        if apply_Kcorr:
            Kcorr = catalog.get_k_correction(band, z_array, color_names[band], colors[i])
        else:
            Kcorr = 0.
        # subtract part where absolute magnitude (if it was located at that z) of the galaxy exceeds Mmax
        subt = 0.
        inds = np.where(M_mdl(ms[i],cosmology.dl_zH0(z_array),Kcorr=Kcorr) > Mmax)[0]
        if inds.size != 0:
            ind = int(inds[-1])
            low_zcut = min(zcut, z_array[ind])
            subt = truncnorm.cdf(low_zcut, low_z_lim, high_z_lim, zs[i], sigmazs[i]) 
        galaxy_norm += truncnorm.cdf(zcut, low_z_lim, high_z_lim, zs[i], sigmazs[i]) - subt

    logger.info(f"nside {nside} pixel {pixel_index}: The effective number of galaxies with z=0 within 3sigma is {N_in3sigma}")
    if len(zs) != len(np.where(zs<zcut)[0]):
        logger.warn(f'The effective number of galaxies below z={zcut} is estimated to be {galaxy_norm}.')

    return galaxy_norm    


def create_norm_map(outfile, catalog, band, nside, pixel_index, zmax, zcut, mth, cosmology, schech_Mmax, apply_Kcorr):
    """
    Generates a healpix map containing an average norm value 
    based on the galaxies of a particular pixel in the sky.
    The result is saved as a fits file.

    Parameters
    ----------
    outfile : str
        path to the output file
    catalog : str
        galaxy catalogue
    band : str
        bbservation band of galaxy catalog (B,K,W1,bJ,u,g,r,i,z)
    nside : int
        healpy skymap nside
    pixel_index: int
        healpy skymap pixel index for which mth is calculated
        if None, a full skymap is generated
    zmax: float
        upper redshift limit for integrals
    zcut: float
        hard redshift cut to apply to the galaxy catalogue
    mth: str
        path to the magnitude threshold map
    cosmology: gwcosmo.utilities.cosmology.standard_cosmology object
        cosmological model
    schech_Mmax : float
        value for Mmax of schechter function for a given band   
    apply_Kcorr: bool
        Apply K-corrections
    
    Returns
    -------
    None
    """

    logger.info(f"create {outfile}")
    galaxy_catalog = gwcosmo.prior.catalog.load_catalog(catalog, band)

    mth_map = hp.fitsfunc.read_map(mth, nest=True)  

    npix = hp.pixelfunc.nside2npix(int(nside))
    m = np.zeros(npix)

    ipix = [pixel_index] # one pixel
    if pixel_index == None:
        ipix = np.arange(0, npix, 1) # full sky

    for pix in ipix:
        mth_val = mth_map[pix]
        galaxy_norm = calc_norm(galaxy_catalog, band, nside, pix, zmax, zcut, mth_val, cosmology, schech_Mmax, apply_Kcorr)
        m[pix] = galaxy_norm

    hp.fitsfunc.write_map(outfile, m, nest=True, overwrite=True)



"""
Line-of-sight redshift prior creation module

Rachel Gray
Freija Beirnaert
"""

import logging
logger = logging.getLogger(__name__)

import numpy as np
from scipy.integrate import quad, dblquad
from scipy.stats import norm, truncnorm
from scipy.interpolate import interp1d
from gwcosmo.utilities.zprior_utilities import create_zarray
from gwcosmo.utilities.luminosity_function import M_Mobs,M_mdl
from gwcosmo.prior.catalog import color_names, color_limits
from gwcosmo.likelihood.skymap import ra_dec_from_ipix, ipix_from_ra_dec
from gwcosmo.maps.create_norm_map import calc_norm
from optparse import make_option
import healpy as hp

import logging
logger = logging.getLogger(__name__)

class LineOfSightRedshiftPrior(object):
    """
    Calculate the likelihood of cosmological parameters from one GW event, 
    using the galaxy catalogue method.

    Parameters
    ----------
    pixel_index : Index of the healpy pixel to analyse
    galaxy_catalog : gwcosmo.prior.catalog.galaxyCatalog object
        The galaxy catalogue
    schechter_params : gwcosmo.utilities.schechter_params.SchechterParams class
        Class that stores the schechter function parameters alpha, Mstar, Mmin, Mmax
    cosmology : gwcosmo.utilities.cosmology.standard_cosmology object
        Cosmological model
    Kcorr : bool, optional
        Should K corrections be applied to the analysis? (default=False)
        Will raise an error if used in conjunction with a galaxy catalogue
        without sufficient color information.
    mth : string, optional
        Read the apparent magnitude threshold from a (lower resolution) precomputed mth map located at this path
        (default=None). If None, mth is estimated from the galaxy catalogue. If "inf", the pixel is empty
    galaxy_norm : string, optional
        Read the galaxy_normalization from a (lower resolution) precomuted galaxy norm map located at this path
        (default=None). If None, galaxy_norm is computed on the fly for this pixel and resolution.
    zcut : float, optional
        An artificial redshift cut to the galaxy catalogue (default=None)
    zmax : float, optional
        The upper redshift limit for integrals (default=10.). Should be well
        beyond the highest redshift reachable by GW data or selection effects.
    zuncert : bool, optional
        Should redshift uncertainties be marginalised over? (Default=True).

    """


    def __init__(self, pixel_index, galaxy_catalog, observation_band, nside, schechter_params, cosmology, zprior, luminosity_prior, luminosity_weights, Kcorr=False, mth=None, galaxy_norm=None, zcut=None, zmax=10., catalog_is_complete=False, min_gals_for_threshold=10):

        """
        Parameters
        ----------
        pixel_index : int
            The healpy index of the pixel being analysed
        galaxy_catalog : objecthttps://git.ligo.org/cbc-cosmo/gwcosmo
            The galaxy catalogue
        observation_band : str
            Observation band (eg. 'B', 'K', 'u', 'g')
        nside : int
            The resolution value of nside
        cosmology : object
            Standard cosmology
        zprior : object
            redshift prior, p(z)
        zrates : object
            rate evolution function, p(s|z)
        luminosity_prior : object
            absolute magnitude prior, p(M|H0)
        luminosity_weights : object
            luminosity weighting function, p(s|M)
        Kcorr : bool, optional
            Should K corrections be applied to the analysis? (default=False)
        mth : string, optional
            Path to the precomputed (lower resolution) apparent magnitude threshold map
        galaxy_norm : string, optional
            Path to the precomputed (lower resolution) galaxy_norm map
        zcut : float, optional
            An artificial redshift cut to the galaxy catalogue (default=None)
        zmax : float, optional
            The upper redshift limit for the universe (default=10.)
        zuncert : bool, optional
            Should redshift uncertainties be marginalised over? (Default=True)
        complete_catalog : bool, optional
            is the galaxy catalogue already complete? (Default=False)
        """
        
        self.pixel_index = pixel_index
        self.nside = nside
        self.zprior = zprior
        self.luminosity_prior = luminosity_prior
        self.luminosity_weights = luminosity_weights
        self.zcut = zcut
        self.sp = schechter_params
        self.Kcorr = Kcorr
        self.band = observation_band
        self.zmax = zmax
        self.cosmo = cosmology
        self.catalog_is_complete = catalog_is_complete
        self.galaxy_catalog = galaxy_catalog
        # load in the galaxy catalogue data from this pixel
        subcatalog = galaxy_catalog.select_pixel(nside, pixel_index, nested=True)
        
        # Set colour limits based on whether Kcorrections are applied
        if Kcorr == True:
            self.color_limit = color_limits[color_names[observation_band]]
        else:
            self.color_limit = [-np.inf,np.inf]

        if mth is None:
            # Calculate mth on the fly
            self.mth = subcatalog.magnitude_thresh(observation_band,min_gals=min_gals_for_threshold)
        elif mth == "inf":
            # For empty catalogue calculation
            self.mth = np.inf
        else:
            # Use precomputed mth map
            # Load precomputed map
            m = hp.fitsfunc.read_map(mth, nest=True)   
            # Get the coordinates of the pixel centre
            pixra, pixdec = ra_dec_from_ipix(nside, pixel_index, nest=True)  
            # Compute the corresponding low-res index
            ipix = ipix_from_ra_dec(hp.pixelfunc.get_nside(m), pixra, pixdec, nest=True)
            # Look up mth in precomputed map
            self.mth = m[ipix]
        logger.info(f"pixel {pixel_index}: mth = {self.mth}")     
        
        # select galaxies which are below the redshift cut and magnitude threshold, and have the relevant color info
        logger.info(f"pixel {pixel_index}: Selecting galaxies below redshift {self.zcut}, with color information between {self.color_limit[0]} and {self.color_limit[1]} in the {observation_band} band, and below apparent magnitude threshold {self.mth}")
        subcatalog = subcatalog.apply_redshift_cut(self.zcut).apply_color_limit(observation_band, self.color_limit[0], self.color_limit[1]).apply_magnitude_limit(observation_band, self.mth)

        self.zs = subcatalog['z'] 
        #ras = subcatalog['ra']
        #decs = subcatalog['dec']
        self.ms = subcatalog.get_magnitudes(observation_band)
        self.sigmazs = subcatalog['sigmaz']
        self.colors = subcatalog.get_color(observation_band)

        self.galaxy_norm = 0.
        if not np.isinf(self.mth):
            if galaxy_norm is None:
                logger.warn("galaxy_norm is None, which should not be used at high resolutions")
                # calculate on the fly 
                galaxy_norm = calc_norm(self.galaxy_catalog, self.band, nside, pixel_index, self.zmax, self.zcut, self.mth, self.cosmo, self.sp.Mmax, self.Kcorr)
                self.galaxy_norm = galaxy_norm
            else:
                # Use precomputed galaxy norm map
                # Load precomputed map
                m = hp.fitsfunc.read_map(galaxy_norm, nest=True)   
                # Get the coordinates of the pixel centre
                pixra, pixdec = ra_dec_from_ipix(nside, pixel_index, nest=True)  
                # Compute the corresponding low-res index
                ipix = ipix_from_ra_dec(hp.pixelfunc.get_nside(m), pixra, pixdec, nest=True)
                # Look up mth in precomputed map
                self.galaxy_norm = m[ipix]
                # Adjust for different resolutions    
                self.galaxy_norm /= (hp.nside2npix(self.nside) / hp.get_map_size(m))
        logger.info(f"galaxy norm: {self.galaxy_norm}")

    def get_ngals(self):
        return len(self.zs)
        

    def uninformative_host_galaxy_prior(self,M,z,H0):
        """
        evaluates p(z)*p(M|H0)*p(s|M)
        
        This is the prior distribution of all GW host galaxies in the universe,
        neglecting rate evolution of mergers with redshift

        Parameters
        ----------
        M : float
            absolute magnitude
        z : float
            redshift
        H0 : float
            Hubble constant value in kms-1Mpc-1

        Returns
        -------
        float
        """
        
        return self.zprior(z)*self.luminosity_prior(M,H0)*self.luminosity_weights(M)


    def uninformative_galaxy_prior(self,M,z,H0):
        """
        evaluates p(z)*p(M|H0)
        
        This is the prior distribution of all galaxies in the universe

        Parameters
        ----------
        M : float
            absolute magnitude
        z : float
            redshift
        H0 : float
            Hubble constant value in kms-1Mpc-1

        Returns
        -------
        float
        """
        return self.zprior(z)*self.luminosity_prior(M,H0)   


    def create_redshift_prior(self):
        logger.info(f"nside {self.nside} pixel {self.pixel_index}: Start creation of redshift prior")
        z_array = create_zarray(self.galaxy_catalog, self.band)
        z_array2 = np.logspace(-6,np.log10(self.zmax),600) # low-res redshift array for smooth functions

        dz = np.ediff1d(z_array,to_end=1)
        
        H0 = self.cosmo.H0
        Mmin = M_Mobs(H0,self.sp.Mmin)
        Mmax = M_Mobs(H0,self.sp.Mmax)

        # if the pixel has no galaxy catalogue support, compute redshift
        if np.isinf(self.mth):
            pz_G = 0
            pG = 0
            pz_Gbar = np.zeros(len(z_array2))
            for i in range(len(z_array2)):
                pz_Gbar[i] = quad(self.uninformative_host_galaxy_prior,Mmin,Mmax,args=(z_array2[i],H0), epsabs=0, epsrel=1.49e-6)[0]
        else:    
            pz_G = np.zeros(len(z_array))
            for i in range(len(self.zs)):
                z_array3 = np.linspace(self.zs[i]-5*self.sigmazs[i],self.zs[i]+5*self.sigmazs[i],50) # redshift array for truncnorm interpolations in redshiftuncertainties
                trunk = np.empty_like(z_array3)
                # set redshift limits so that galaxies can't have negative z
                low_z_lim, high_z_lim = (0 - self.zs[i])/self.sigmazs[i], (self.zmax - self.zs[i])/self.sigmazs[i]
                if self.Kcorr:
                    Kcorr = self.galaxy_catalog.get_k_correction(self.band, z_array, color_names[self.band], self.colors[i])
                else:
                    Kcorr = 0.
                trunk = truncnorm.pdf(z_array3, low_z_lim, high_z_lim, self.zs[i], self.sigmazs[i])
                interpolate_trunk = interp1d(z_array3, trunk, bounds_error=False, fill_value=0) 
                kill_weights = M_mdl(self.ms[i],self.cosmo.dl_zH0(z_array),Kcorr=Kcorr) < Mmax
                pz_G += interpolate_trunk(z_array) * self.luminosity_weights(M_mdl(self.ms[i],self.cosmo.dl_zH0(z_array),Kcorr=Kcorr)) * kill_weights
            if self.galaxy_norm != 0:
                pz_G = pz_G/self.galaxy_norm
                pz_G[np.where(z_array>self.zcut)[0]]=0
            else:
                pz_G = 0
            
            pG = dblquad(self.uninformative_galaxy_prior,0,self.zcut,lambda x: Mmin, lambda x: max(min(M_mdl(self.mth,self.cosmo.dl_zH0(x)),Mmax),Mmin), args=[H0], epsabs=0, epsrel=1.49e-6)[0]

            # compute p(z|Gbar,H0)
            pz_Gbar = np.zeros(len(z_array2))
            for i in range(len(z_array2)):
                if z_array2[i] < self.zcut:
                    pz_Gbar[i] = quad(self.uninformative_host_galaxy_prior, min(max(M_mdl(self.mth,self.cosmo.dl_zH0(z_array2[i])),Mmin),Mmax), Mmax, args=(z_array2[i],H0), epsabs=0, epsrel=1.49e-6)[0]
                else:
                    pz_Gbar[i] = quad(self.uninformative_host_galaxy_prior, Mmin, Mmax, args=(z_array2[i],H0), epsabs=0, epsrel=1.49e-6)[0]
            
        pz_Gbar_smooth = interp1d(z_array2,pz_Gbar,kind='linear')

        pz_total = (pz_G*pG + pz_Gbar_smooth(z_array))

        return pz_total, z_array



import numpy as np
import astropy.constants as const


def sph2vec(ra, dec):
    return np.array([np.sin(np.pi/2.-dec)*np.cos(ra),np.sin(np.pi/2.-dec)*np.sin(ra),np.cos(np.pi/2.-dec)])

def zhelio_to_zcmb(ra, dec, z_helio):
    speed_of_light = const.c.to('km/s').value
    ra_cmb = 167.99
    dec_cmb = -7.22
    v_cmb = 369.
    z_gal_cmb = v_cmb*np.dot(sph2vec(ra_cmb,dec_cmb),sph2vec(ra,dec))/speed_of_light
    return (1.+z_helio)*(1.+z_gal_cmb)-1.


"""
Module for managing Injections and calculate selection effects.

Benoit Revenu, Christos Karathanasis, Simone Mastrogiovanni
"""

import numpy as _np
import copy as _copy
import sys as _sys
import h5py as _h5py
import pickle as _pickle
import bilby as _bilby
from astropy import units as _u
from scipy.special import logsumexp as _logsumexp


__all__=['injections_at_detector','injections_at_source']


class Injections():
    
    def __init__(self,m1d,m2d,dl,prior_vals,snr_det,snr_cut,ifar,ifar_cut,ntotal,Tobs,condition_check=False,Nobs=-1):

        """
        This class is used to manage a list of detected injections to calculated
        GW selection effects. This class uses injections which are given in source frame.

        Parameters
        ----------
        m1d: _np.arrray
            Mass 1 detector frame of detected events (provide if file_injection is not provided)
        m2d: _np.arrray
            Mass 2 detector frame of detected events (provide if file_injection is not provided)
        dl: _np.arrray
            luminosity distance of detected events (provide if file_injection is not provided)
        prior_vals: _np.arrray
            Used prior draws for injections (provide if file_injection is not provided). This array contains the injections probabilities either
            in the source frame or detector frame, according to the child class that initializes it (either injections_at_detector or injections_at_source)
        snr_det: _np.arrray
            SNR of detected events (provide if file_injection is not provided)
        snr_cut: float
            Set different to 0 if you wanto to apply a different SNR cut.
        ifar: _np.array
            IFAR of the detected events (IFAR = Inverse False Alarm Rate)
        ifar_cut: float
            Set different to 0 if you wanto to apply a different IFAR cut.
        ntotal: float
            Total number of simulated injections (detected and not). This is necessary to compute the expected number of detections
        Tobs: float
            Length of time for the run in years (used to calculate rates)
        condition_check: boolean, True or False
            is True, we check wether the injections cover the entire prior range
        Nobs: integer
            it's the number of observed GW events in the current analysis
        """
        # Saves what you provided in the class
        self.condition_check = condition_check
        self.snr_cut = snr_cut
        self.ntotal = 1.*ntotal # floating value to prevent an overflow when computing ntotal**2
        self.ntotal_original = 1.*ntotal # floating value
        self.dl_original = dl
        self.m1d_original = m1d
        self.m2d_original = m2d
        self.snr_original = snr_det
        self.ini_prior_original = prior_vals
        self.ifar = ifar
        self.ifar_cut = ifar_cut
        self.Tobs = Tobs
        self.Nobs = Nobs # needed for the check on Neff >= 4*Nobs, the check will give True in the default case (Nobs==-1)

        idet=_np.where((self.snr_original>snr_cut) & (self.ifar>self.ifar_cut))[0]

        self.idet=idet
        self.m1det=m1d[idet]
        self.m2det=m2d[idet]
        self.dldet=dl[idet]
        self.snrdet=self.snr_original[idet]
        self.ini_prior=self.ini_prior_original[idet]

    def update_cut(self,snr_cut=0,ifar_cut=0,fraction=None):
        print('Selecting injections with SNR {:f} and IFAR {:f} yr'.format(snr_cut,ifar_cut))

        self.snr_cut=snr_cut
        self.ifar_cut=ifar_cut
        idet=_np.where((self.snr_original>snr_cut) & (self.ifar>self.ifar_cut))[0]

        #Sub-sample the selected injections in order to reduce the computational load
        if fraction is not None and (fraction > 0) and (fraction <= 1):
            idet=_np.random.choice(idet,size=int(len(idet)*fraction),replace=False)
            self.ntotal=int(self.ntotal_original*fraction)
            print('Working with a total of {:d} injections'.format(len(idet)))

        self.idet=idet
        self.m1det=self.m1d_original[idet]
        self.m2det=self.m2d_original[idet]
        self.dldet=self.dl_original[idet]
        self.snrdet=self.snr_original[idet]
        self.ini_prior=self.ini_prior_original[idet]
        self.log_jacobian_det=self.log_origin_jacobian[idet]

    
    def detector_frame_to_source_frame(self,cosmo,m1det,m2det,dldet):
        
        z_samples = cosmo.z_dgw(dldet)

        ms1 = m1det/(1+z_samples)
        ms2 = m2det/(1+z_samples)

        return ms1, ms2, z_samples


    def detector_to_source_jacobian(self,z,cosmo):
        
        jacobian = _np.power(1+z,2)*cosmo.ddgw_dz(z)

        return jacobian

    
    def update_VT(self,cosmo,m_prior,z_prior,z_prior_norm):
        """
        This method updates the sensitivity estimations.

        Parameters
        ----------
        m_prior : mass prior class
            mass prior class from the prior.mass module
        z_prior : redshift prior class
        """

        self.ms1, self.ms2, self.z_samples = self.detector_frame_to_source_frame(cosmo,self.m1det,self.m2det,self.dldet)

        # Checks if the injections cover the entire prior range. If not, throws an errror if the flag is True
        if self.condition_check:
            mass_a = _np.hstack([ms1,ms2])
            if (_np.max(mass_a)<m_prior.mmax) | (_np.min(mass_a)>m_prior.mmin):
                print('The  injections source frame masses are not enough to cover all the prior range')
                print('Masses prior range {:.2f}-{:.2f} Msol'.format(m_prior.mmin,m_prior.mmax))
                print('Injections range {:.2f}-{:.2f} Msol'.format(_np.min(mass_a),_np.max(mass_a)))
                _sys.exit()


        log_numer = _np.log(z_prior(self.z_samples))+m_prior.log_joint_prob(self.ms1,self.ms2)
        log_jacobian_term = _np.log(_np.abs(self.detector_to_source_jacobian(self.z_samples, cosmo)))-self.log_jacobian_det
        self.log_weights_astro = log_numer-_np.log(self.ini_prior)-log_jacobian_term
        self.log_weights = self.log_weights_astro - _np.log(z_prior_norm)
        # This is the Volume-Time in which we expect to detect. You can multiply it by R_0 Tobs to get the expected number of detections in Gpc^3 yr
        self.VT_sens=_np.exp(_logsumexp(self.log_weights_astro))/self.ntotal
        # This is the fraction of events we expect to detect, a.k.a. the selection effect
        self.VT_fraction=self.VT_sens/z_prior_norm

    def calculate_Neff(self):
        """
        Calculates the effective number of injections contributing to the calculation of the selection effect.
        returns Neff and the boolean result of (Neff>= 4*self.Nobs) and the variance
        See Eq 9 of (https://arxiv.org/pdf/1904.10879.pdf) for more details
        """
        mean = self.VT_fraction
        var = _np.exp(_logsumexp(self.log_weights*2))/(self.ntotal**2)-(mean**2)/self.ntotal
        Neff = (mean**2)/var
        return Neff, (Neff >= 4*self.Nobs), var

    def expected_number_detection(self,R0):
        """
        This method will return the expected number of GW detection given the injection set. Tobs is automatically saved in the class creation

        Parameters
        ----------
        R0 : float
            Merger rate in comoving volume in Gpc-3yr-1
        """
        return self.VT_sens*R0*self.Tobs

    def gw_only_selection_effect(self):
        """
        Will evaluate the GW-only selection effect using the set of injections

        Returns
        -------
        Selection effect (float)
        """
        return self.VT_fraction

    
class injections_at_detector(Injections):
    """
    A class to handle a list of detected GW signals from simulations in detectors frame. This can be used to
    evaluate selection effects or detection expectations under some priors
    """
    
    def __init__(self,m1d,m2d,dl,prior_vals,snr_det,snr_cut,ifar,ifar_cut,ntotal,Tobs,condition_check=False):
        
        self.log_origin_jacobian = _np.zeros(len(m1d))
        Injections.__init__(self,m1d,m2d,dl,prior_vals,snr_det,snr_cut,ifar,ifar_cut,ntotal,Tobs,condition_check=False)

    
class injections_at_source(Injections):
    """
    A class to handle a list of detected GW signals from simulations in source frame. This can be used to
    evaluate selection effects or detection expectations under some priors
    """

    def __init__(self,cosmo_ref,m1s,m2s,z,prior_vals,snr_det,snr_cut,ifar,ifar_cut,ntotal,Tobs,condition_check=False):

        self.cosmo_ref = cosmo_ref
        self.z_original = z
        self.m1s_original = m1s
        self.m2s_original = m2s
        
        # Convert from source frame to detector frame and select injections according to SNR and IFAR
        m1d, m2d, dl = self.source_frame_to_detector_frame(self.cosmo_ref,self.m1s_original,self.m2s_original,self.z_original)
        self.log_origin_jacobian = _np.log(_np.abs(self.detector_to_source_jacobian(self.z_original,self.cosmo_ref)))
        Injections.__init__(self,m1d,m2d,dl,prior_vals,snr_det,snr_cut,ifar,ifar_cut,ntotal,Tobs,condition_check=False)
        
        
    def source_frame_to_detector_frame(self,cosmo,ms_1,ms_2,z_samples):
        
        dl = cosmo.dgw_z(self, z_samples)

        md1 = ms_1*(1+z_samples)
        md2 = ms_2*(1+z_samples)

        return md1, md2, dl
